# Load packages
from bs4 import BeautifulSoup
from dotenv import load_dotenv
from IPython.display import display, HTML, Markdown
import json
import os
import requests

# Declare function
def FetchWebsiteText(url,
                     browserless_api_key=None):
    """
    Fetches the text content from a specified website using the Browserless service.

    Parameters:
        url (str): The URL of the website from which to fetch text.
        browserless_api_key (str, optional): The API key for the Browserless service. If not provided, the function attempts to load it from a .env file.

    Returns:
        str: The text content of the website, with any occurrence of four or more newlines replaced with three newlines.
    """
    
    # If no API key is provided, try to load from .env file
    if browserless_api_key is None:
        load_dotenv()
        try:
            browserless_api_key = os.environ['BROWERLESS_API_KEY']
        except:
            raise ValueError("No API key provided and no .env file found. If you need a Browserless API key, visit https://www.browserless.io/")
    
    # Set payload to query
    payload = json.dumps({
        'url': url,
    })
    
    # Define headers for the API call
    headers = {
        'Cache-Control': 'no-cache',
        'Content-Type': 'application/json'
    }
    
    # Send the API call to browserless
    post_url = "https://chrome.browserless.io/content?token=" + browserless_api_key
    response = requests.request(
        method="POST", 
        url=post_url,
        headers=headers,
        data=payload
    )
    
    # Use BeautifulSoup to parse the response
    if response.status_code == 200:
        soup = BeautifulSoup(response.text, 'html.parser')
        text = soup.get_text()
    else:
        raise ValueError("Error fetching website contents. Status code: " + str(response.status_code))
    
    # Replace 4 or more newlines with 3 newlines
    text = re.sub(r'\n{4,}', '\n\n\n', text)
    
    # Return the text
    return text




# Load packages
import numpy as np
import pandas as pd

# Declare function
def ConvertOddsToProbability(dataframe,
                             odds_column,
                             probability_column_name=None):
    """
    This function converts odds to probability.

    Args:
        dataframe (Pandas dataframe): Pandas dataframe containing the data to be analyzed.
        odds_column (str): The name of the column containing the odds.
        probability_column_name (str, optional): The name of the column for the probability. Defaults to None.

    Returns:
        Pandas dataframe: Dataframe with odds converted to probability in a new column.
    """
    
    # If probability column name is not specified, set it to odds column name + "- as probability"
    if probability_column_name is None:
        probability_column_name = odds_column + " - as probability"
    
    # Convert odds to probability
    dataframe[probability_column_name] = np.where(
        (dataframe[odds_column].isnull()) | (dataframe[odds_column]+1 == 0),
        np.nan,
        dataframe[odds_column] / (1 + dataframe[odds_column])
    )
    
    # Return updated dataframe
    return dataframe




from .CalculateEigenvalues import CalculateEigenvalues
from .ConvertMatrixToRowEchelonForm import ConvertMatrixToRowEchelonForm
from .ConvertSystemOfEquationsToMatrix import ConvertSystemOfEquationsToMatrix
from .PlotVectors import PlotVectors
from .SolveSystemOfEquations import SolveSystemOfEquations
from .VisualizeMatrixAsLinearTransformation import VisualizeMatrixAsLinearTransformation


# Load packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def PlotFunction(f_of_x, 
                 minimum_x=-10, 
                 maximum_x=10, 
                 n=100,
                 # Line formatting arguments
                 line_color="#3269a8",
                 line_alpha=0.8,
                 color_palette="Set2",
                 markers="o",
                 # Text formatting arguments
                 x_axis_variable_name="x",
                 y_axis_variable_name="f(x)",
                 title_for_plot="Function Plot",
                 subtitle_for_plot="Shows the function of x",
                 caption_for_plot=None,
                 data_source_for_plot=None,
                 x_indent=-0.127,
                 title_y_indent=1.125,
                 subtitle_y_indent=1.05,
                 caption_y_indent=-0.3,
                 # Plot formatting arguments
                 figure_size=(8, 5)):
    """
    This function plots a function of x.

    Args:
        f_of_x (lambda): The function of x to plot.
        minimum_x (int, optional): The minimum value of x to use when plotting the function. Defaults to -10.
        maximum_x (int, optional): The maximum value of x to use when plotting the function. Defaults to 10.
        n (int, optional): The number of points to use when plotting the function. Defaults to 100.
        line_color (str, optional): The color of the function plot line. Defaults to "#3269a8".
        line_alpha (float, optional): The alpha (transparency) of the function plot line. Defaults to 0.8.
        color_palette (str, optional): The color palette to use for the markers. Defaults to "Set2".
        markers (str, optional): The marker style to use for the function plot. Defaults to "o".
        x_axis_variable_name (str, optional): The label for the x-axis. Defaults to "x".
        y_axis_variable_name (str, optional): The label for the y-axis. Defaults to "f(x)".
        title_for_plot (str, optional): The title for the plot. Defaults to "Function Plot".
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to "Shows the function of x".
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        x_indent (float, optional): The x-indent for the x-axis label. Defaults to -0.127.
        title_y_indent (float, optional): The y-indent for the title. Defaults to 1.125.
        subtitle_y_indent (float, optional): The y-indent for the subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent for the caption. Defaults to -0.3.
        figure_size (tuple, optional): The size of the plot figure. Defaults to (8, 5).
    """
    
    # Plot the function
    x = np.linspace(minimum_x, maximum_x, n)
    y = np.zeros(len(x))
    for i in range(len(x)):
        y[i] = f_of_x(x[i])
    
    # Convert the function to a dataframe
    dataframe = pd.DataFrame({
        "x": x,
        "y": y
    })
        
    # Create figure and axes
    fig, ax = plt.subplots(figsize=figure_size)
    
    # Create line plot with seaborn
    ax = sns.lineplot(
        data=dataframe,
        x=x,
        y=y,
        color=line_color,
        alpha=line_alpha,
        marker=markers,
        ax=ax
    )
    
    # Remove top and right spines, and set bottom and left spines to gray
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_color('#666666')
    ax.spines['left'].set_color('#666666')

    # Format tick labels to be Arial, size 9, and color #666666
    ax.tick_params(
        which='major',
        labelsize=9,
        color='#666666'
    )
    
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Word wrap the subtitle without splitting words
    if subtitle_for_plot != None:   
        subtitle_for_plot = textwrap.fill(subtitle_for_plot, 100, break_long_words=False)
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
    
    # Move the y-axis label to the top of the y-axis, and set the font to Arial, size 9, and color #666666
    ax.yaxis.set_label_coords(-0.1, 0.84)
    ax.yaxis.set_label_text(
        y_axis_variable_name,
        fontname="Arial",
        fontsize=10,
        color="#666666"
    )
    
    # Move the x-axis label to the right of the x-axis, and set the font to Arial, size 9, and color #666666
    ax.xaxis.set_label_coords(0.9, -0.1)
    ax.xaxis.set_label_text(
        x_axis_variable_name,
        fontname="Arial",
        fontsize=10,
        color="#666666"
    )
    
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )

    # Show plot
    plt.show()




# Load packages
import os

# Declare function
def CreateFileTree(path, 
                   indent_spaces=2):
    """
    Recursively walks the directory tree and prints a diagram of all the subdirectories and files in the tree.

    Args:
        path (str): The path to the root directory.
        indent (str): The number of spaces to indent the output.
    """
    
    # Initialize the output
    output = []
    
    # Walk the directory tree
    for root, dirs, files in os.walk(path):
        # Get the current directory
        for dir in dirs:
            output.append(f'{dir}')
            # Get the files in the current directory
            for file in os.listdir(os.path.join(root, dir)):
                # Indent the files
                output.append(f'├{"─" * indent_spaces}{file}')

    # Print the output
    print('\n'.join(output))




# Load packages
import pandas as pd
import os
from math import ceil
from matplotlib import pyplot as plt
import seaborn as sns
import textwrap

# Declare function
def PlotRiskTolerance(simulated_values,
                      risk_tolerance,
                      # Risk tolerance arguments
                      lower_is_worse=True,
                      variable_name="Outcome",
                      observed_value=None,
                      risk_tolerance_label="Tolerance",
                      # Histogram formatting arguments
                      fill_color="#999999",
                      fill_transparency=0.6,
                      # Text formatting arguments
                      title_for_plot="Risk Tolerance",
                      subtitle_for_plot="Shows the simulated outcomes that are worse than the risk tolerance.",
                      caption_for_plot=None,
                      data_source_for_plot=None,
                      show_y_axis=False,
                      title_y_indent=1.1,
                      subtitle_y_indent=1.05,
                      caption_y_indent=-0.15,
                      # Plot formatting arguments
                      figure_size=(8, 6)):
    """
    This function plots a histogram of simulated values and highlights the values that are worse than the risk tolerance.
    
    Args:
        simulated_values (numpy array or pandas Series): The simulated values to plot.
        risk_tolerance (int or float): The risk tolerance value.
        lower_is_worse (bool, optional): If True, values lower than the risk tolerance are considered worse. Defaults to True.
        variable_name (str, optional): The name of the variable being simulated. Defaults to "Outcome".
        observed_value (int or float, optional): The observed value to be plotted. Defaults to None.
        risk_tolerance_label (str, optional): The label for the risk tolerance. Defaults to "Tolerance".
        fill_color (str, optional): The fill color for the histogram. Defaults to "#999999".
        fill_transparency (float, optional): The transparency of the fill color. Defaults to 0.6.
        show_mean (bool, optional): If True, the mean of the simulated values is shown. Defaults to True.
        show_median (bool, optional): If True, the median of the simulated values is shown. Defaults to True.
        title_for_plot (str, optional): The title for the plot. Defaults to "Risk Tolerance".
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to "Show the simulated outcomes that are worse than the risk tolerance.".
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        show_y_axis (bool, optional): If False, the y-axis is not shown. Defaults to False.
        title_y_indent (float, optional): The y-indent for the title. Defaults to 1.1.
        subtitle_y_indent (float, optional): The y-indent for the subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent for the caption. Defaults to -0.15.
        figure_size (tuple, optional): The size of the figure. Defaults to (8, 6).
    """
    
    # Ensure that risk tolerance is numeric
    if not isinstance(risk_tolerance, (int, float)):
        raise ValueError("Risk tolerance must be numeric.")
    
    # Ensure that observed_value is numeric if it is provided
    if observed_value != None and not isinstance(observed_value, (int, float)):
        raise ValueError("Observed value must be numeric.")
    
    # If simulated_values is a pandas series or a list, convert it to a numpy array
    if isinstance(simulated_values, pd.Series) or isinstance(simulated_values, list):
        simulated_values = simulated_values.to_numpy()
    
    # Create dataframe with only the outcome variable and the simulated_values
    dataframe = pd.DataFrame(simulated_values, columns=[variable_name])
    
    # Create risk flag column name
    if lower_is_worse:
        risk_flag_col_name = "Is Below Risk Tolerance"
    else:
        risk_flag_col_name = "Is Above Risk Tolerance"
    
    # Add flag for whether the outcome is worse than the risk tolerance
    if lower_is_worse:
        dataframe[risk_flag_col_name] = dataframe[variable_name] < risk_tolerance
    else:
        dataframe[risk_flag_col_name] = dataframe[variable_name] > risk_tolerance
        
    # Create figure and axes
    fig, ax = plt.subplots(figsize=figure_size)
    
    # Create histogram using seaborn
    sns.histplot(
        data=dataframe,
        x=variable_name,
        color=fill_color,
        alpha=fill_transparency,
        hue=risk_flag_col_name,
        palette={True: "#FF0000", False: "#999999"}
    )
    
    # Remove top, left, and right spines. Set bottom spine to dark gray.
    ax.spines['top'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_color("#262626")
    
    # Remove the y-axis, and adjust the indent of the plot titles
    if show_y_axis == False:
        ax.axes.get_yaxis().set_visible(False)
        x_indent = 0.015
    else:
        x_indent = -0.005
    
    # Remove the y-axis label
    ax.set_ylabel(None)
    
    # Calculate the percentage of outcomes that are worse than the risk tolerance
    risk_probability = dataframe[risk_flag_col_name].mean()
    
    # Show the risk tolerance as a vertical line with a label and probability
    ax.axvline(
        x=risk_tolerance,
        ymax=0.97-.02,
        color="#262626",
        linestyle="--",
        linewidth=1.5,
        alpha=0.5
    )
    ax.text(
        x=risk_tolerance, 
        y=plt.ylim()[1] * 0.97, 
        s=risk_tolerance_label+': {:.2f}'.format(risk_tolerance)+' ({:.1%})'.format(risk_probability),
        horizontalalignment='center',
        # fontname="Arial",
        fontsize=9,
        color="#262626",
        alpha=0.75
    )
    
    # Show the oberseved value as a vertical line with a label
    if observed_value != None:
        ax.axvline(
            x=observed_value,
            ymax=0.90-.02,
            color="#262626",
            linewidth=2,
            alpha=0.8
        )
        ax.text(
            x=observed_value, 
            y=plt.ylim()[1] * .90,
            s=variable_name+': {:.2f}'.format(observed_value),
            horizontalalignment='center',
            # fontname="Arial",
            fontsize=9,
            color="#262626",
            alpha=0.75
        )
    
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
    
    # Set x-axis tick label font to Arial, size 9, and color #666666
    ax.tick_params(
        axis='x',
        which='major',
        labelsize=9,
        labelcolor="#666666",
        pad=2,
        bottom=True,
        labelbottom=True
    )
    # plt.xticks(fontname='Arial')
    
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )
        
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()
    



# Turn off warnings
import warnings
warnings.filterwarnings('ignore')

# Import all modules
from .calculus import *
from .data_collection import *
from .data_processing import *
from .descriptive_analytics import *
from .file_management import *
from .hypothesis_testing import *
from .linear_algebra import *
from .predictive_analytics import *
from .prescriptive_analytics import *
from .simulations import *
from .visualizations import *

# Turn on warnings
warnings.filterwarnings('default')


# Load packages
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def PlotHeatmap(dataframe,
                x_axis_column_name,
                y_axis_column_name,
                value_column_name,
                # Heatmap formatting arguments
                color_map_minimum_value=None,
                color_map_minimum_color=13,
                color_map_maximum_value=None,
                color_map_maximum_color=125,
                color_map_center_value=None,
                color_map_saturation=80,
                color_map_lightness=60,
                show_legend=False,
                figure_size=(8, 6),
                border_line_width=6,
                border_line_color="#ffffff",
                square_cells=False,
                # Text formatting arguments
                show_data_labels=True,
                data_label_fontweight="normal",
                data_label_color="#262626",
                data_label_fontsize=12,
                data_label_format="d",
                title_for_plot=None,
                subtitle_for_plot=None,
                caption_for_plot=None,
                data_source_for_plot=None,
                title_y_indent=1.15,
                subtitle_y_indent=1.1,
                caption_y_indent=-0.17):
    # Convert dataframe to wide-form dataframe, using x_axis_column_name as the index and y_axis_column_name as the columns
    dataframe = dataframe.pivot(
        index=x_axis_column_name, 
        columns=y_axis_column_name,
        values=value_column_name
    )
    
    # Transpose dataframe
    dataframe = dataframe.transpose()
    
    # Create figure and axes
    fig, ax = plt.subplots(figsize=figure_size)
    
    # Plot contingency table in a heatmap using seaborn
    ax = sns.heatmap(
        data=dataframe,
        cmap=sns.diverging_palette(
            color_map_minimum_color, 
            color_map_maximum_color, 
            s=color_map_saturation, 
            l=color_map_lightness,
            as_cmap=True
        ),
        center=color_map_center_value,
        vmin=color_map_minimum_value,
        vmax=color_map_maximum_value,
        annot=show_data_labels,
        fmt=data_label_format,
        linewidths=border_line_width,
        linecolor=border_line_color,
        cbar=show_legend,
        square=square_cells,
        ax=ax,
        # Bold the text in the heatmap
        annot_kws={"fontweight": data_label_fontweight,
                #    "fontname": "Arial", 
                   "fontsize": data_label_fontsize,
                   "color": data_label_color}
    )
    
    # Wrap y axis label using textwrap
    wrapped_variable_name = "\n".join(textwrap.wrap(y_axis_column_name, 30))  # String wrap the variable name
    ax.set_ylabel(
        wrapped_variable_name, 
        fontsize=12, 
        # fontname="Arial", 
        color="#262626"
    )
    
    # Format and wrap y axis tick labels using textwrap
    y_tick_labels = ax.get_yticklabels()
    wrapped_y_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 40, break_long_words=False)) for label in y_tick_labels]
    ax.set_yticklabels(
        wrapped_y_tick_labels, 
        fontsize=10, 
        # fontname="Arial", 
        color="#262626"
    )
    
    # Wrap x axis label using textwrap
    wrapped_variable_name = "\n".join(textwrap.wrap(x_axis_column_name, 30))  # String wrap the variable name
    ax.set_xlabel(
        wrapped_variable_name, 
        fontsize=12, 
        # fontname="Arial", 
        color="#262626"
    )
    
    # Format and wrap x axis tick labels using textwrap
    x_tick_labels = ax.get_xticklabels()
    wrapped_x_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 40, break_long_words=False)) for label in x_tick_labels]
    ax.set_xticklabels(
        wrapped_x_tick_labels, 
        fontsize=10, 
        # fontname="Arial", 
        color="#262626"
    )
    
    # Set the x indent of the plot titles and captions
    # Get longest y tick label
    longest_y_tick_label = max(wrapped_y_tick_labels, key=len)
    if len(longest_y_tick_label) >= 30:
        x_indent = -0.3
    else:
        x_indent = -0.005 - (len(longest_y_tick_label) * 0.011)
        
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
    
        
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )
        
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages
import sys
import os
import win32com.client as client

# Declare function
def ConvertWordDocsToPDF(word_folder_path,
                         pdf_folder_path,
                         open_each_doc=False):
    """
    This function converts all Word documents in a folder to PDF.

    Args:
        word_folder_path (str): The path to the folder containing the Word documents.
        pdf_folder_path (str): The path to the folder where the PDF documents will be saved.
        open_each_doc (bool, optional): Whether to open each Word document as it is converted. Defaults to False.
    """
    
    # Create list of word files in the directory
    list_word_docs = [f for f in os.listdir(word_folder_path) if f.endswith(".docx")]
    
    # Connect to Microsoft Word
    word = client.Dispatch('Word.Application')
    
    # Open the word document without opening the application, unless open_each_doc is True
    if open_each_doc == False:
        word.Visible = False
    else:
        word.Visible = True
    
    # Convert each word file to PDF
    for word_doc in list_word_docs:
        # Create full path to Word document
        word_doc_path = os.path.join(word_folder_path, word_doc)
        # Create full path to PDF document
        pdf_doc_path = os.path.join(pdf_folder_path, word_doc.replace(".docx", ".pdf"))
        print("Converting {} to PDF...".format(word_doc))
        # Try to save the document as PDF
        current_doc = word.Documents.Open(word_doc_path)
        try:
            current_doc.SaveAs(pdf_doc_path, FileFormat=17)
        except Exception as e:
            # Create PDF folder if it doesn't exist
            os.makedirs(name=pdf_folder_path)
            current_doc.SaveAs(pdf_doc_path, FileFormat=17)
        current_doc.Close()
    word.Quit()
        
    # Print message saying script is done
    print("All Word documents converted to PDF.")




# Load packages
import io
import requests
import zipfile

# Declare function
def GetZipFile(url,
               path_to_save_folder,
               unzip=True,
               print_contents=True):
    """
    This function downloads a zip file from a url and saves it to a specified folder. 
    It can also unzip the file and print the contents of the zip file.
    
    Args:
        url (str): The url of the zip file to download.
        path_to_save_folder (str): The path to the folder where the zip file will be saved.
        unzip (bool, optional): Whether or not to unzip the file. Defaults to True.
        print_contents (bool, optional): Whether or not to print the contents of the zip file. Defaults to True.
    """
    
    # Download the zip file from the url
    r = requests.get(url)
    z = zipfile.ZipFile(io.BytesIO(r.content))
    
    # Unzip the file
    if unzip:
        z.extractall(path_to_save_folder)
    
    # Print contents of the zip file
    if print_contents:
        print("Contents of the zip file:")
        for item in z.namelist():
            print(item)




# Load packages
from IPython.display import display, Markdown
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn import metrics
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor, plot_tree, export_text
import textwrap

# Declare function
def CreateDecisionTreeModel(dataframe,
                            outcome_variable,
                            list_of_predictor_variables,
                            is_outcome_categorical=True,
                            # Model training arguments
                            test_size=0.2,
                            categorical_splitting_criterion='entropy',
                            numerical_splitting_criterion='mse',
                            maximum_depth=None,
                            minimum_impurity_decrease=0.0,
                            random_seed=412,
                            filter_nulls=False,
                            # All plot arguments
                            data_source_for_plot=None,
                            # Model performance plot arguments
                            plot_model_test_performance=True,
                            dot_fill_color="#999999",
                            fitted_line_type=None,
                            line_color=None,
                            heatmap_color_palette="Blues",
                            figure_size_for_model_test_performance_plot=(8, 6),
                            title_for_model_test_performance_plot="Model Performance",
                            subtitle_for_model_test_performance_plot="The predicted values vs. the actual values in the test dataset.",
                            caption_for_model_test_performance_plot=None,
                            title_y_indent_for_model_test_performance_plot=1.09,
                            subtitle_y_indent_for_model_test_performance_plot=1.05,
                            caption_y_indent_for_model_test_performance_plot=-0.215,
                            x_indent_for_model_test_performance_plot=-0.115,
                            # Feature importance plot arguments
                            plot_feature_importance=True,
                            top_n_to_highlight=3,
                            highlight_color="#b0170c",
                            fill_transparency=0.8,
                            figure_size_for_feature_importance_plot=(8, 6),
                            title_for_feature_importance_plot="Feature Importance",
                            subtitle_for_feature_importance_plot="Shows the predictive power of each feature in the model.",
                            caption_for_feature_importance_plot=None,
                            title_y_indent_for_feature_importance_plot=1.15,
                            subtitle_y_indent_for_feature_importance_plot=1.1,
                            caption_y_indent_for_feature_importance_plot=-0.15,
                            # Decision tree plot arguments
                            plot_decision_tree=False,
                            decision_tree_plot_size=(20, 20),
                            print_decision_rules=False):
    # Keep only the predictors and outcome variable
    dataframe = dataframe[list_of_predictor_variables + [outcome_variable]].copy()
    
    # Drop rows with infinite values
    dataframe = dataframe.replace([np.inf, -np.inf], np.nan)
    
    # Drop rows with missing values if filter_nulls is True
    if filter_nulls:
        dataframe = dataframe.dropna()
    print("Count of examples eligible for inclusion in model training and testing:", len(dataframe.index))
    
    # Split dataframe into training and test sets
    train, test = train_test_split(
        dataframe, 
        test_size=test_size,
        random_state=random_seed
    )
    
    # Create decision tree model
    if is_outcome_categorical:
        model = DecisionTreeClassifier(
            criterion=categorical_splitting_criterion,
            max_depth=maximum_depth,
            min_impurity_decrease=minimum_impurity_decrease,
            random_state=random_seed
        )
    else:
        model = DecisionTreeRegressor(
            criterion=numerical_splitting_criterion,
            max_depth=maximum_depth,
            min_impurity_decrease=minimum_impurity_decrease,
            random_state=random_seed
        )
        
    # Fit the model
    model = model.fit(train[list_of_predictor_variables], train[outcome_variable])
    
    # Add predictions to test set
    test['Predicted'] = model.predict(test[list_of_predictor_variables])
    
    # Show mean squared error and variance if outcome is numerical
    if is_outcome_categorical == False:
        print('Mean Squared Error:', metrics.mean_squared_error(test[outcome_variable], test['Predicted']))
        print('Variance Score:', metrics.r2_score(test[outcome_variable], test['Predicted']))
        print("Note: A variance score of 1 is perfect prediction and 0 means that there is no linear relationship between X and Y.")
    # Show accuracy if outcome is categorical
    else:
        classifcation_report = metrics.classification_report(test[outcome_variable], test['Predicted'])
        print("Classification Report:\n", classifcation_report, sep="")
    
    # Print decision rules if requested
    if print_decision_rules:
        r = export_text(
            model,
            feature_names=list_of_predictor_variables
        )
        print(r)
    
    # Plot decision tree if requested
    if plot_decision_tree:
        plt.figure(figsize=decision_tree_plot_size)
        plot_tree(
            model, 
            feature_names=list_of_predictor_variables,
            filled=True,
            rounded=True,
            precision=3
        )
    
    # Print the confusion matrix if outcome is categorical
    if plot_model_test_performance:
        # Set the size of the plot
        plt.figure(figsize=figure_size_for_model_test_performance_plot)
        
        # If outcome is categorical, create a heatmap
        if is_outcome_categorical:
            # Generate a contingency table using pandas
            contingency_table = pd.crosstab(
                test['Predicted'],
                test[outcome_variable],
                rownames=['Predicted'],
                colnames=['Actual'],
                margins=True,
                margins_name='Total',
                dropna=False
            )
            
            # Display as a markdown table
            display(contingency_table)
            
            # Create a confusion matrix
            confusion_matrix = metrics.confusion_matrix(
                test[outcome_variable], 
                test['Predicted']
            )
            
            # Transpose the confusion matrix
            confusion_matrix = confusion_matrix.transpose()
            
            # Convert the confusion matrix to a percentage using the column sums
            confusion_matrix = confusion_matrix / confusion_matrix.sum(axis=0)
            
            # Generate a heatmap of the confusion matrix
            plt.figure(figsize=(9,9))
            ax = sns.heatmap(
                confusion_matrix, 
                annot=True, 
                # Format numbers as percentages
                fmt='.0%', 
                linewidths=.5, 
                square=True, 
                cmap=heatmap_color_palette,
            )
            
            # Remove the color bar
            ax.collections[0].colorbar.remove()

        # Plot the residuals if outcome is numerical
        else:
            # Generate a scatterplot of the predicted vs. observed outcome
            ax = sns.regplot(
                data=test,
                x=outcome_variable,
                y='Predicted',
                marker='o',
                scatter_kws={
                    'color': dot_fill_color,
                    'alpha': 0.5,
                    'linewidth': 0.5,
                    'edgecolor': dot_fill_color
                },
                lowess=True,
                line_kws={'color': line_color}
            )
            
            # Add a "perfect prediction" line
            plt.plot(
                test[outcome_variable], 
                test[outcome_variable], 
                color='black', 
                alpha=0.35, 
                linewidth=0.5, 
                linestyle='--'
            )
            
        # Remove top and right spines, and set bottom and left spines to gray
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color('#666666')
        ax.spines['left'].set_color('#666666')
        
        # Format tick labels to be Arial, size 9, and color #666666
        ax.tick_params(
            which='major',
            labelsize=9,
            color='#666666'
        )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent_for_model_test_performance_plot,
            y=title_y_indent_for_model_test_performance_plot,
            s=title_for_model_test_performance_plot,
            # fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent_for_model_test_performance_plot,
            y=subtitle_y_indent_for_model_test_performance_plot,
            s=subtitle_for_model_test_performance_plot,
            # fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Move the y-axis label to the top of the y-axis, and set the font to Arial, size 9, and color #666666
        ax.yaxis.set_label_coords(-0.1, 0.92)
        ax.yaxis.set_label_text(
            'Predicted',
            # fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
        # Move the x-axis label to the right of the x-axis, and set the font to Arial, size 9, and color #666666
        ax.xaxis.set_label_coords(0.9, -0.1)
        ax.xaxis.set_label_text(
            outcome_variable,
            # fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
        # Add a word-wrapped caption if one is provided
        if caption_for_model_test_performance_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_model_test_performance_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_model_test_performance_plot, 130, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent_for_model_test_performance_plot,
                y=caption_y_indent_for_model_test_performance_plot,
                s=wrapped_caption,
                # fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
            # Show the plot
            plt.show()
    
    # Plot feature importance if requested
    if plot_feature_importance:
        # Create dataframe of feature importance
        data_feauture_importance = pd.DataFrame(
            data={
                'Feature': model.feature_names_in_,
                'Importance': model.feature_importances_
            }
        )
        
        # Sort dataframe by importance
        data_feauture_importance = data_feauture_importance.sort_values(by='Importance', ascending=False)
        
        # Highlight top n features
        data_feauture_importance['Highlighted'] = np.where(
            data_feauture_importance['Feature'].isin(data_feauture_importance['Feature'].head(top_n_to_highlight)),
            True,
            False
        )
        
        # Plot feature importance with seaborn, using a horizontal barplot
        plt.figure(figsize=figure_size_for_feature_importance_plot)
        ax = sns.barplot(
            data=data_feauture_importance,
            x='Importance',
            y='Feature',
            hue='Highlighted',
            palette={True: highlight_color, False: "#b8b8b8"},
            alpha=fill_transparency,
            dodge=False
        )
        
        # Remove the legend
        ax.legend_.remove()
        
        # Format and wrap y axis tick labels using textwrap
        y_tick_labels = ax.get_yticklabels()
        wrapped_y_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 50)) for label in y_tick_labels]
        ax.set_yticklabels(
            wrapped_y_tick_labels, 
            fontsize=10, 
            # fontname="Arial", 
            color="#262626"
        )
        
        # Remove a-axis tick labels
        ax.get_xaxis().set_ticks([])
        
        # Format x-axis label
        ax.set_xlabel(
            "Importance", 
            fontsize=10, 
            # fontname="Arial", 
            color="#262626"
        )
        
        # Remove spines
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_color('#b8b8b8')
        ax.spines['bottom'].set_visible(False)
        
        # Add data labels
        for container in ax.containers:
            ax.bar_label(
                container, 
                fmt='%.3f', 
                label_type='edge', 
                padding=5,
                fontsize=10, 
                # fontname="Arial", 
                color="#262626"
            )
        
        # Add space between the title and the plot
        plt.subplots_adjust(top=0.85)
        
        # Set the x indent of the plot titles and captions
        # Get longest y tick label
        longest_y_tick_label = max(wrapped_y_tick_labels, key=len)
        if len(longest_y_tick_label) >= 30:
            x_indent = -0.3
        else:
            x_indent = -0.005 - (len(longest_y_tick_label) * 0.011)
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent_for_feature_importance_plot,
            s=title_for_feature_importance_plot,
            # fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent_for_feature_importance_plot,
            s=subtitle_for_feature_importance_plot,
            # fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Add a word-wrapped caption if one is provided
        if caption_for_feature_importance_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_feature_importance_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_feature_importance_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent_for_feature_importance_plot,
                s=wrapped_caption,
                # fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show the plot
        plt.show()
        plt.clf()
     
    # Return the model
    return model




# Load packages
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import norm
import seaborn as sns
import textwrap

# Declare function
def ConductAnomalyDetection(dataframe, 
                            list_of_columns_to_analyze,
                            anomaly_threshold=0.95,
                            plot_detection_summary=True,
                            summary_plot_size=(20, 20),
                            column_name_for_anomaly_prob='Anomaly Probability',
                            column_name_for_anomaly_flag='Anomaly Detected'):
    """
    This function conducts anomaly detection on a dataset using the z-score method.
    
    Args:
        dataframe (Pandas dataframe): Pandas dataframe containing the data to be analyzed.
        list_of_columns_to_analyze (list): List of predictor variables to be analyzed.
        anomaly_threshold (float, optional): _description_. The threshold for the probability of an anomaly. Defaults to 0.95.
        plot_detection_summary (bool, optional): _description_. Whether to plot a summary of the anomaly detection. Defaults to True.
        summary_plot_size (tuple, optional): _description_. The size of the summary plot. Defaults to (20, 20).
        column_name_for_anomaly_prob (str, optional): _description_. The name of the column for the anomaly probability. Defaults to 'Anomaly Probability'.
        
    Returns:
        Pandas dataframe: Pandas dataframe containing the data to be analyzed with the anomaly probability and flag.
    """
    
    # If column_name_for_anomaly_prob is in the dataframe, drop it
    if column_name_for_anomaly_prob in dataframe.columns:
        dataframe = dataframe.drop(column_name_for_anomaly_prob, axis=1)
        
    # If column_name_for_anomaly_flag is in the dataframe, drop it
    if column_name_for_anomaly_flag in dataframe.columns:
        dataframe = dataframe.drop(column_name_for_anomaly_flag, axis=1)
    
    # Keep only the predictor variables
    dataframe_anomaly = dataframe[list_of_columns_to_analyze].copy()
    
    # Keep complete cases
    dataframe_anomaly = dataframe_anomaly.replace([np.inf, -np.inf], np.nan)
    dataframe_anomaly = dataframe_anomaly.dropna()
    
    # Get z-score from probability
    z_score = norm.ppf(anomaly_threshold / 2)
    
    # Iterate through each predictor variable
    dataframe_anomaly[column_name_for_anomaly_prob] = 1
    for predictor_variable in list_of_columns_to_analyze:
        # Get the mean and standard deviation of the predictor variable
        variable_mean = dataframe_anomaly[predictor_variable].mean()
        variable_std = dataframe_anomaly[predictor_variable].std()
        # Get z-score for each observation
        dataframe_anomaly[predictor_variable + ' z-score'] = (dataframe_anomaly[predictor_variable] - variable_mean) / variable_std
        # Update probability of anomaly
        dataframe_anomaly[column_name_for_anomaly_prob] *= norm.cdf(abs(dataframe_anomaly[predictor_variable + ' z-score']))
        # Drop the z-score column
        dataframe_anomaly = dataframe_anomaly.drop(predictor_variable + ' z-score', axis=1)
    
    # Join anomaly probability to original dataset
    dataframe = dataframe.merge(
        dataframe_anomaly[[column_name_for_anomaly_prob]],
        how='left', 
        left_index=True, 
        right_index=True
    )
    
    # Add flag for anomaly if probability is below threshold
    dataframe[column_name_for_anomaly_flag] = np.where(
        dataframe[column_name_for_anomaly_prob] > anomaly_threshold,
        True,
        False
    )
    
    # Show pairplot of the data
    if plot_detection_summary:
        # Generate a pairplot of the data
        plt.figure(figsize=summary_plot_size)
        sns.pairplot(
            data=dataframe[list_of_columns_to_analyze + [column_name_for_anomaly_flag]],
            hue=column_name_for_anomaly_flag
        )
        
        # Word wrap the axis labels
        for ax in plt.gcf().axes:
            ax.set_xlabel(textwrap.fill(ax.get_xlabel(), 40))
            ax.set_ylabel(textwrap.fill(ax.get_ylabel(), 40))
        
        # Show the plot
        plt.show()
    
    # Return the dataframe
    return(dataframe)




from .AddDateNumberColumns import AddDateNumberColumns
from .AddLeadingZeros import AddLeadingZeros
from .AddRowCountColumn import AddRowCountColumn
from .AddTPeriodColumn import AddTPeriodColumn
from .AddTukeyOutlierColumn import AddTukeyOutlierColumn
from .CleanTextColumns import CleanTextColumns
from .ConductAnomalyDetection import ConductAnomalyDetection
from .ConductEntityMatching import ConductEntityMatching
from .ConvertOddsToProbability import ConvertOddsToProbability
from .CountMissingDataByGroup import CountMissingDataByGroup
from .CreateBinnedColumn import CreateBinnedColumn
from .CreateDataOverview import CreateDataOverview
from .CreateRandomSampleGroups import CreateRandomSampleGroups
from .CreateRareCategoryColumn import CreateRareCategoryColumn
from .CreateStratifiedRandomSampleGroups import CreateStratifiedRandomSampleGroups
from .ImputeMissingValuesUsingNearestNeighbors import ImputeMissingValuesUsingNearestNeighbors
from .VerifyGranularity import VerifyGranularity


# Load packages
import numpy as np
import pandas as pd
from sklearn.impute import KNNImputer

# Declare function
def ImputeMissingValuesUsingNearestNeighbors(dataframe,
                                             list_of_numeric_columns_to_impute,
                                             number_of_neighbors=3,
                                             averaging_method='uniform'):
    """
    This function imputes missing values in a dataframe using nearest neighbors.
    
    Args:
        dataframe (Pandas dataframe): Pandas dataframe
        list_of_numeric_columns_to_impute (list): The list of variables to impute.
        number_of_neighbors (int, optional): The number of neighbors to use for imputation. Defaults to 3.
        averaging_method (str, optional): The weight function used in prediction. Defaults to 'uniform'.
    
    Returns:
        Pandas dataframe: An updated Pandas dataframe with imputed values.
    """
    
    # Select only the variables to impute
    dataframe_imputed = dataframe[list_of_numeric_columns_to_impute].copy()
    
    # Impute missing values using nearest neighbors
    imputer = KNNImputer(n_neighbors=number_of_neighbors,
                         weights=averaging_method)
    
    # Fit the imputer
    dataframe_imputed = imputer.fit_transform(dataframe_imputed)
    
    # Add "- Imputed" to the variable names
    list_new_column_names = []
    for variable in list_of_numeric_columns_to_impute:
        variable_imputed = variable + " - Imputed"
        list_new_column_names.append(variable_imputed)
    
    # Convert the imputed array to a dataframe
    dataframe_imputed = pd.DataFrame(dataframe_imputed,
                                     columns=list_new_column_names)
    
    # Bind the imputed dataframe to the original dataframe
    dataframe = pd.concat([dataframe, dataframe_imputed[list_new_column_names]], axis=1)
    
    # Return the dataframe with imputed values
    return(dataframe)




# Load pacakges
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn import linear_model, metrics
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
import textwrap

# Declare function
def CreateLinearRegressionModel(dataframe,
                                outcome_variable,
                                list_of_predictor_variables,
                                # Model parameters
                                scale_variables=False,
                                test_size=0.2,
                                fit_intercept=True,
                                random_seed=412,
                                print_peak_to_peak_range_of_each_predictor=False,
                                # All plot arguments
                                data_source_for_plot=None,
                                # Model performance plot arguments
                                plot_model_test_performance=True,
                                dot_fill_color="#999999",
                                line_color=None,
                                figure_size_for_model_test_performance_plot=(8, 6),
                                title_for_model_test_performance_plot="Model Performance",
                                subtitle_for_model_test_performance_plot="The predicted values vs. the actual values in the test dataset.",
                                caption_for_model_test_performance_plot=None,
                                title_y_indent_for_model_test_performance_plot=1.10,
                                subtitle_y_indent_for_model_test_performance_plot=1.05,
                                caption_y_indent_for_model_test_performance_plot=-0.215,
                                x_indent_for_model_test_performance_plot=-0.115,
                                # Feature importance plot arguments
                                plot_feature_importance=True,
                                top_n_to_highlight=3,
                                highlight_color="#b0170c",
                                fill_transparency=0.8,
                                figure_size_for_feature_importance_plot=(8, 6),
                                title_for_feature_importance_plot="Feature Importance",
                                subtitle_for_feature_importance_plot="Shows the predictive power of each feature in the model.",
                                caption_for_feature_importance_plot=None,
                                title_y_indent_for_feature_importance_plot=1.15,
                                subtitle_y_indent_for_feature_importance_plot=1.1,
                                caption_y_indent_for_feature_importance_plot=-0.15):
    # Keep only the predictors and outcome variable
    dataframe = dataframe[list_of_predictor_variables + [outcome_variable]].copy()
    
    # Replace inf with nan, and drop rows with nan
    dataframe.replace([np.inf, -np.inf], np.nan, inplace=True)
    dataframe.dropna(inplace=True)
    print("Count of examples eligible for inclusion in model training and testing:", len(dataframe.index))
    
    # Scale the predictors, if requested
    if scale_variables:
        # Scale predictors
        scaler = StandardScaler()
        dataframe[list_of_predictor_variables] = scaler.fit_transform(dataframe[list_of_predictor_variables])
        
    # Show the peak-to-peak range of each predictor
    if print_peak_to_peak_range_of_each_predictor:
        print("\nPeak-to-peak range of each predictor:")
        print(np.ptp(dataframe[list_of_predictor_variables], axis=0))
    
    # Split dataframe into training and test sets
    if test_size > 0:
        train, test = train_test_split(
            dataframe, 
            test_size=test_size,
            random_state=random_seed
        )
    else:
        train = dataframe.copy()
        test = dataframe.copy()
        
    # Create linear regression object
    if scale_variables:
        model = make_pipeline(StandardScaler(),
                              linear_model.LinearRegression(
                                  fit_intercept=fit_intercept,
                              )
        )
    else:
        model = linear_model.LinearRegression(
            fit_intercept=fit_intercept,
        )
    
    # Train the model using the training sets and show fitting summary
    model.fit(X=train[list_of_predictor_variables], 
              y=train[outcome_variable])
    
    # Show number of iterations and weight updates
    if scale_variables:
        regressor = model['linearregression']
    else:
        regressor = model
    
    # # Show parameters of the model
    # b_norm = model.intercept_
    # w_norm = model.coef_
    # print(f"\nModel parameters:    w: {w_norm}, b:{b_norm}")
    
    # Add predictions to test set
    test['Predicted'] = model.predict(test[list_of_predictor_variables])
    
    # Show mean squared error if outcome is numerical
    print('Mean Squared Error:', metrics.mean_squared_error(test[outcome_variable], test['Predicted']))
    print('Variance Score:', metrics.r2_score(test[outcome_variable], test['Predicted']))
    print("Note: A variance score of 1 is perfect prediction and 0 means that there is no linear relationship between X and Y.")
    
    # Plot predicted and observed outputs if requested
    if plot_model_test_performance:
        # Set the size of the plot
        plt.figure(figsize=figure_size_for_model_test_performance_plot)
        
        # Generate a scatterplot of the predicted vs. observed outcome
        ax = sns.regplot(
            data=test,
            x=outcome_variable,
            y='Predicted',
            marker='o',
            scatter_kws={
                'color': dot_fill_color,
                'alpha': 0.5,
                'linewidth': 0.5,
                'edgecolor': dot_fill_color
            },
            lowess=True,
            line_kws={'color': line_color}
        )
        
        # Add a "perfect prediction" line
        plt.plot(
            test[outcome_variable], 
            test[outcome_variable], 
            color='black', 
            alpha=0.35, 
            linewidth=0.5, 
            linestyle='--'
        )
        
        # Remove top and right spines, and set bottom and left spines to gray
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color('#666666')
        ax.spines['left'].set_color('#666666')
        
        # Format tick labels to be Arial, size 9, and color #666666
        ax.tick_params(
            which='major',
            labelsize=9,
            color='#666666'
        )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent_for_model_test_performance_plot,
            y=title_y_indent_for_model_test_performance_plot,
            s=title_for_model_test_performance_plot,
            # fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent_for_model_test_performance_plot,
            y=subtitle_y_indent_for_model_test_performance_plot,
            s=subtitle_for_model_test_performance_plot,
            # fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Move the y-axis label to the top of the y-axis, and set the font to Arial, size 9, and color #666666
        ax.yaxis.set_label_coords(-0.1, 0.92)
        ax.yaxis.set_label_text(
            'Predicted',
            # fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
        # Move the x-axis label to the right of the x-axis, and set the font to Arial, size 9, and color #666666
        ax.xaxis.set_label_coords(0.9, -0.1)
        ax.xaxis.set_label_text(
            outcome_variable,
            # fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
        # Add a word-wrapped caption if one is provided
        if caption_for_model_test_performance_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_model_test_performance_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_model_test_performance_plot, 130, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent_for_model_test_performance_plot,
                y=caption_y_indent_for_model_test_performance_plot,
                s=wrapped_caption,
                # fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
            # Show the plot
            plt.show()
    
    # Plot feature importance if requested
    if plot_feature_importance:
        # Get importance
        importance = regressor.coef_
        
        # Create lists to store feature names and feature importance
        feature_names = []
        feature_importance = []
        
        # Store feature names and feature importance in lists
        for i,v in enumerate(importance):
            feature_names.append(i)
            feature_importance.append(v)
        # Create dataframe of feature importance
        data_feauture_importance = pd.DataFrame(
            data={
                'Feature': model.feature_names_in_,
                'Importance': regressor.coef_
            }
        )
        
        # Sort dataframe by importance
        data_feauture_importance = data_feauture_importance.sort_values(by='Importance', ascending=False)
        
        # Highlight top n features
        data_feauture_importance['Highlighted'] = np.where(
            data_feauture_importance['Feature'].isin(data_feauture_importance['Feature'].head(top_n_to_highlight)),
            True,
            False
        )
        
        # Plot feature importance with seaborn, using a horizontal barplot
        plt.figure(figsize=figure_size_for_feature_importance_plot)
        ax = sns.barplot(
            data=data_feauture_importance,
            x='Importance',
            y='Feature',
            hue='Highlighted',
            palette={True: highlight_color, False: "#b8b8b8"},
            alpha=fill_transparency,
            dodge=False
        )
        
        # Remove the legend
        ax.legend_.remove()
        
        # Format and wrap y axis tick labels using textwrap
        y_tick_labels = ax.get_yticklabels()
        wrapped_y_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 50)) for label in y_tick_labels]
        ax.set_yticklabels(
            wrapped_y_tick_labels, 
            fontsize=10, 
            # fontname="Arial", 
            color="#262626"
        )
        
        # Remove a-axis tick labels
        ax.get_xaxis().set_ticks([])
        
        # Format x-axis label
        ax.set_xlabel(
            "Beta Coefficent", 
            fontsize=10, 
            # fontname="Arial", 
            color="#262626"
        )
        
        # Remove spines
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_color('#b8b8b8')
        ax.spines['bottom'].set_visible(False)
        
        # Add data labels
        for container in ax.containers:
            ax.bar_label(
                container, 
                fmt='%.3f', 
                label_type='edge', 
                padding=5,
                fontsize=10, 
                # fontname="Arial", 
                color="#262626"
            )
        
        # Add space between the title and the plot
        plt.subplots_adjust(top=0.85)
        
        # Set the x indent of the plot titles and captions
        # Get longest y tick label
        longest_y_tick_label = max(wrapped_y_tick_labels, key=len)
        if len(longest_y_tick_label) >= 30:
            x_indent = -0.3
        else:
            x_indent = -0.005 - (len(longest_y_tick_label) * 0.011)
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent_for_feature_importance_plot,
            s=title_for_feature_importance_plot,
            # fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent_for_feature_importance_plot,
            s=subtitle_for_feature_importance_plot,
            # fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Add a word-wrapped caption if one is provided
        if caption_for_feature_importance_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_feature_importance_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_feature_importance_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent_for_feature_importance_plot,
                s=wrapped_caption,
                # fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show the plot
        plt.show()
        plt.clf()
    
    # Return the model
    return model




# Import packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import StandardScaler

# Declare function
def CreateGaussianMixtureClusters(dataframe,
                                  list_of_numeric_columns_for_clustering=None,
                                  number_of_clusters=None,
                                  column_name_for_clusters='Gaussian Mixture Cluster',
                                  scale_predictor_variables=False,
                                  show_cluster_summary_plots=True,
                                  sns_color_palette='Set1',
                                  summary_plot_size=(20, 20),
                                  random_seed=412,
                                  maximum_iterations=300):
    """
    This function creates Gaussian Mixture clusters on a dataset based on the variables specified.

    Args:
        dataframe (Pandas dataframe): Pandas dataframe containing the data to be analyzed.
        list_of_numeric_columns_for_clustering (list, optional): The list of variables to base the clusters on. Defaults to None, which will use all variables in the dataframe.
        number_of_clusters (int, optional): The number of clusters to create. Defaults to None, which will use the elbow method to determine the optimal number of clusters.
        column_name_for_clusters (str, optional): The name of the new column containing the clusters. Defaults to 'K-Means Cluster'.
        scale_predictor_variables (bool, optional): Whether to scale the predictor variables prior to analysis. Defaults to True.
        show_cluster_summary_plots (bool, optional): Whether to show cluster summary plots. Defaults to True.
        summary_plot_size (tuple, optional): The size of the summary plots. Defaults to (20, 20).
        random_seed (int, optional): The random seed to use for replication. Defaults to 412.
        maximum_iterations (int, optional): The maximum number of iterations to use for the K-Means algorithm. Defaults to 300.
    
    Returns:
        Pandas dataframe: An updated Pandas dataframe with the clusters joined to the original data.
    """
    
    # Keep complete cases only
    dataframe_clusters = dataframe.dropna(subset=list_of_numeric_columns_for_clustering)

    # Scale the predictors, if requested
    if scale_predictor_variables:
        # Scale predictors
        dataframe_clusters[list_of_numeric_columns_for_clustering] = StandardScaler().fit_transform(dataframe_clusters[list_of_numeric_columns_for_clustering])
    
    # # Show peak-to-peak range of each predictor
    # print("\nPeak-to-peak range of each predictor:")
    # print(np.ptp(dataframe_clusters[list_of_numeric_columns_for_clustering], axis=0))
    
    # If number_of_clusters is None, conduct clustering up to 12 times and plot the results
    if number_of_clusters is None:
        n_clusters = np.arange(1,13)
        models = [GaussianMixture(n, random_state=random_seed).fit(dataframe_clusters[list_of_numeric_columns_for_clustering]) for n in n_clusters]
        plt.plot(
            n_clusters,
            [m.bic(dataframe_clusters) for m in models],
            label = 'BIC'
        )
        plt.plot(
            n_clusters,
            [m.aic(dataframe_clusters) for m in models],
            label = 'AIC'
        )
        plt.legend()
        plt.xlabel('Number of Clusters')
        
        # Find the optimal number of clusters based on the BIC and AIC
        number_of_clusters = np.argmin([m.bic(dataframe_clusters) for m in models]) + 1
        print("\nNumber of clusters based on minimum BIC: " + str(number_of_clusters))
        print("But, be sure to check the AIC and BIC curves on the plot to ensure the optimal number of clusters makes sense for your purpose.")
        
    # Conduct Guassian Mixture clustering
    model = GaussianMixture(
        n_components=number_of_clusters, 
        random_state=random_seed,
        max_iter=maximum_iterations
    )
    model = model.fit(dataframe_clusters[list_of_numeric_columns_for_clustering])

    # Get clusters 
    clusters = pd.Series(model.predict(dataframe_clusters[list_of_numeric_columns_for_clustering]))
    dataframe_clusters[column_name_for_clusters] = clusters

    # Convert cluster to categorical data type
    dataframe_clusters[column_name_for_clusters] = dataframe_clusters[column_name_for_clusters].apply(str)
    
    # Get cluster probabilities 
    probabilities = round(pd.DataFrame(model.predict_proba(dataframe_clusters[list_of_numeric_columns_for_clustering])), 4)
    probabilities.columns = [column_name_for_clusters + '_' + str(i) + ' Probability' for i in range(0, number_of_clusters)]
    dataframe_clusters = pd.concat([dataframe_clusters, probabilities], axis = 1)
    
    # Join clusters to original dataset
    dataframe_clusters = dataframe_clusters[[column_name_for_clusters] + probabilities.columns.tolist()]
    dataframe = dataframe.merge(
        dataframe_clusters,
        how='left',
        left_index=True,
        right_index=True
    )
    
    # If requested, show box plots of each cluster for each variable selected
    if show_cluster_summary_plots:
        plt.figure(figsize=summary_plot_size)
        sns.pairplot(
            data=dataframe[list_of_numeric_columns_for_clustering + [column_name_for_clusters]],
            hue=column_name_for_clusters,
            palette=sns_color_palette,
        )
        plt.suptitle("Cluster Summary Plots", fontsize=15)
        plt.show()

    # Return updated dataset
    return(dataframe)




# Load packages
import pandas as pd
import os
from math import ceil
from matplotlib import pyplot as plt
import seaborn as sns
import textwrap

# Declare function
def PlotSingleVariableCountPlot(dataframe,
                                categorical_column_name, 
                                # Plot formatting arguments
                                color_palette="Set1",
                                fill_color=None,
                                top_n_to_highlight=None,
                                highlight_color="#b0170c",
                                fill_transparency=0.8,
                                add_rare_category_line=False,
                                rare_category_line_color='#b5b3b3',
                                rare_category_threshold=0.05,
                                figure_size=(8, 6),
                                # Text formatting arguments
                                data_label_fontsize=10,
                                title_for_plot=None,
                                subtitle_for_plot=None,
                                caption_for_plot=None,
                                data_source_for_plot=None,
                                title_y_indent=1.15,
                                subtitle_y_indent=1.1,
                                caption_y_indent=-0.15):
    """
    This function creates a bar chart for a single categorical variable. 
    The function can be used to create a bar chart with a single color, or a bar chart with a different color for the top n categories. 
    The function can also be used to add a line to the bar chart to indicate the threshold for rare categories.

    Args:
        dataframe (_type_): The dataframe containing the categorical variable.
        categorical_column_name (_type_): The categorical variable to plot.
        color_palette (str, optional): The Seaborn color palette to use for the bar chart. Defaults to "Set1".
        fill_color (_type_, optional): The color to use for the bar chart. Defaults to None.
        top_n_to_highlight (_type_, optional): The "top N" categories to highlight. Defaults to None.
        highlight_color (str, optional): The color to use for the highlighted "top N" categories. Defaults to "#b0170c".
        fill_transparency (float, optional): The transparency to use for the bar chart. Defaults to 0.8.
        data_label_fontsize (int, optional): The font size to use for the data labels. Defaults to 10.
        title_for_plot (str, optional): The title for the plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to None.
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        title_y_indent (float, optional): The vertical indent for the title. Defaults to 1.15.
        subtitle_y_indent (float, optional): The vertical indent for the subtitle. Defaults to 1.1.
        caption_y_indent (float, optional): The vertical indent for the caption. Defaults to -0.15.
        add_rare_category_line (bool, optional): Whether to add a line to the bar chart to indicate the threshold for rare categories. Defaults to False.
        rare_category_line_color (str, optional): The color to use for the rare category line. Defaults to '#b5b3b3'.
        rare_category_threshold (float, optional): The threshold for rare categories. Defaults to 0.05.
        figure_size (tuple, optional): The size of the figure. Defaults to (8, 6).
    """
    
    # Check that the column exists in the dataframe.
    if categorical_column_name not in dataframe.columns:
        raise ValueError("Column {} does not exist in dataframe.".format(categorical_column_name))
    
    # Ensure that rare category threshold is between 0 and 1
    if rare_category_threshold < 0 or rare_category_threshold > 1:
        raise ValueError("Rare category threshold must be between 0 and 1.")
    
    # Make sure that the top_n_to_highlight is a positive integer, and less than the number of categories
    if top_n_to_highlight != None:
        if top_n_to_highlight < 0 or top_n_to_highlight > len(dataframe[categorical_column_name].value_counts()):
            raise ValueError("top_n_to_highlight must be a positive integer, and less than the number of categories.")
    
    # Create figure and axes
    fig, ax = plt.subplots(figsize=figure_size)

    # Generate bar chart
    if top_n_to_highlight != None:
        ax = sns.countplot(
            data=dataframe,
            y=categorical_column_name,
            order=dataframe[categorical_column_name].value_counts(ascending=False).index,
            palette=["#b8b8b8" if x not in dataframe[categorical_column_name].value_counts().index[:top_n_to_highlight] else highlight_color for x in dataframe[categorical_column_name].value_counts().index],
            alpha=fill_transparency
        )
    elif fill_color == None:
        ax = sns.countplot(
            data=dataframe,
            y=categorical_column_name,
            order=dataframe[categorical_column_name].value_counts(ascending=False).index,
            palette=color_palette,
            alpha=fill_transparency
        )
    else:
        ax = sns.countplot(
            data=dataframe,
            y=categorical_column_name,
            order=dataframe[categorical_column_name].value_counts(ascending=False).index,
            color=fill_color,
            alpha=fill_transparency
        )
    
    # Add space between the title and the plot
    plt.subplots_adjust(top=0.85)
    
    # Wrap y axis label using textwrap
    wrapped_variable_name = "\n".join(textwrap.wrap(categorical_column_name, 40, break_long_words=False))  # String wrap the variable name
    ax.set_ylabel(wrapped_variable_name)
    
    # Format and wrap y axis tick labels using textwrap
    y_tick_labels = ax.get_yticklabels()
    wrapped_y_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 40, break_long_words=False)) for label in y_tick_labels]
    ax.set_yticklabels(
        wrapped_y_tick_labels, 
        fontsize=10, 
        # fontname="Arial", 
        color="#262626"
    )
    
    # Move x-axis to the top
    ax.xaxis.tick_top()
    
    # Change x-axis colors to "#666666"
    ax.tick_params(axis='x', colors="#666666")
    ax.spines['top'].set_color("#666666")
    
    # Set x-axis title to "Count"
    ax.set_xlabel(
        "Count", 
        fontsize=10, 
        # fontname="Arial", 
        color="#262626"
    )
    
    # Remove bottom, left, and right spines
    ax.spines['bottom'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    
    # Add data labels
    abs_values = dataframe[categorical_column_name].value_counts(ascending=False)
    rel_values = dataframe[categorical_column_name].value_counts(ascending=False, normalize=True).values * 100
    lbls = [f'{p[0]} ({p[1]:.0f}%)' for p in zip(abs_values, rel_values)]
    ax.bar_label(container=ax.containers[0],
                 labels=lbls,
                 padding=5,
                 fontsize=data_label_fontsize)
    
    # Add rare category threshold line
    if add_rare_category_line:
        ax.axvline(
            x=rare_category_threshold * dataframe.shape[0],
            color=rare_category_line_color,
            alpha=0.5,
            linestyle='--',
            label='Rare category threshold'
        )
        
    # Set the x indent of the plot titles and captions
    # Get longest y tick label
    longest_y_tick_label = max(wrapped_y_tick_labels, key=len)
    if len(longest_y_tick_label) >= 30:
        x_indent = -0.3
    else:
        x_indent = -0.005 - (len(longest_y_tick_label) * 0.011)
        
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
      
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )
        
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages
import pandas as pd

# Define function
def AddDateNumberColumns(dataframe,
                         date_column_name):
    """
    This function adds year, quarter, month, and week number columns to a dataset based on a date column.
    
    Args:
        dataframe (Pandas dataframe): Pandas dataframe
        date_column_name (str): Name of column containing dates

    Returns:
        Pandas dataframe: An updated Pandas dataframe with date number columns.
    """
    
    # Extract year from date
    new_column_name = date_column_name + '.Year'
    dataframe[new_column_name] = pd.DatetimeIndex(dataframe[date_column_name]).year

    # Extract quarter from date
    new_column_name = date_column_name + '.Quarter'
    dataframe[new_column_name] = pd.DatetimeIndex(dataframe[date_column_name]).quarter

    # Extract month number from date
    new_column_name = date_column_name + '.Month'
    dataframe[new_column_name] = pd.DatetimeIndex(dataframe[date_column_name]).month

    # Extract week number from date
    new_column_name = date_column_name + '.Week'
    dataframe[new_column_name] = pd.DatetimeIndex(dataframe[date_column_name]).week

    # Extract day from date
    new_column_name = date_column_name + '.Day'
    dataframe[new_column_name] = pd.DatetimeIndex(dataframe[date_column_name]).day
    
    # Extract day of the week
    new_column_name = date_column_name + '.DayOfWeek'
    dataframe[new_column_name] = pd.DatetimeIndex(dataframe[date_column_name]).dayofweek
    print("Note: .DayOfWeek is 0-based starting on Monday (i.e. 0 = Monday, 6 = Sunday).")
    
    # Return dataframe
    return(dataframe)




# Load packages
from math import ceil
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import ttest_1samp
import seaborn as sns
import textwrap

# Declare function
def OneSampleTTest(dataframe, 
                   value_column, 
                   hypothesized_value,
                   confidence_interval=.95, 
                   alternative_hypothesis="two-sided",
                   null_handling='omit',
                   plot_sample_distribution=True,
                   # Plotting arguments
                   fill_color="#999999",
                   fill_transparency=0.2,
                   title_for_plot="One Sample T-Test",
                   subtitle_for_plot="Shows the distribution of the sample mean and the hypothesized value.",
                   caption_for_plot=None,
                   data_source_for_plot=None,
                   show_y_axis=False,
                   title_y_indent=1.10,
                   subtitle_y_indent=1.05,
                   caption_y_indent=-0.15,
                   figure_size=(8, 6)):
    """
    Conducts a one-sample t-test on a given dataframe and value column, comparing the sample mean to a hypothesized value.
    
    Args:
        dataframe (pandas.DataFrame): The dataframe containing the data to be tested.
        value_column (str): The name of the column in the dataframe containing the data to be tested.
        hypothesized_value (float): The hypothesized value to be tested against the sample mean.
        confidence_interval (float, optional): The confidence level for the test. Defaults to 0.95.
        alternative_hypothesis (str, optional): The alternative hypothesis for the test. Can be "two-sided", "less", or "greater". Defaults to "two-sided".
        null_handling (str, optional): How to handle null values in the data. Can be "omit", "raise", or "propagate". Defaults to "omit".
        plot_sample_distribution (bool, optional): Whether to plot the sample distribution and the hypothesized value. Defaults to True.
        fill_color (str, optional): The fill color for the sample distribution plot. Defaults to "#999999".
        fill_transparency (float, optional): The fill transparency for the sample distribution plot. Defaults to 0.2.
        title_for_plot (str, optional): The title for the sample distribution plot. Defaults to "One Sample T-Test".
        subtitle_for_plot (str, optional): The subtitle for the sample distribution plot. Defaults to "Shows the distribution of the sample mean and the hypothesized value.".
        caption_for_plot (str, optional): The caption for the sample distribution plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the sample distribution plot. Defaults to None.
        show_y_axis (bool, optional): Whether to show the y-axis on the sample distribution plot. Defaults to False.
        title_y_indent (float, optional): The y-indent for the title on the sample distribution plot. Defaults to 1.10.
        subtitle_y_indent (float, optional): The y-indent for the subtitle on the sample distribution plot. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent for the caption on the sample distribution plot. Defaults to -0.15.
        figure_size (tuple, optional): The size of the sample distribution plot. Defaults to (8, 6).
    
    Returns:
        scipy.stats.Ttest_1sampResult: The results of the one-sample t-test.
    """
    
    # Conduct one sample t-test
    ttest_results = ttest_1samp(
        dataframe[value_column], 
        popmean=hypothesized_value, 
        axis=0, 
        nan_policy=null_handling, 
        alternative=alternative_hypothesis
    )
    
    # Calculate sample size
    sample_size = len(dataframe[value_column])
    
    # Calculate degrees of freedom
    degrees_of_freedom = sample_size - 1
    
    # Calculate the sample mean
    sample_mean = dataframe[value_column].mean()
    
    # Calculate the standard error of the mean
    standard_error_of_the_mean = dataframe[value_column].sem()
    
    # Get the t-statistic
    t_statistic = ttest_results.statistic
    
    # Get the p-value
    p_value = ttest_results.pvalue
    
    # Plot the sample distribution and the hypothesized value
    if plot_sample_distribution:
        # Generate an array of sample means
        sample_means = np.random.normal(
            loc=sample_mean, 
            scale=standard_error_of_the_mean, 
            size=10000
        )
        
        # Convert array to dataframe
        sample_means = pd.DataFrame(sample_means,
                                    columns=[value_column])
        
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
        
        # Create histogram using seaborn
        ax = sns.kdeplot(
            data=sample_means,
            x=value_column,
            color=fill_color,
            alpha=fill_transparency,
            fill=True,
            common_norm=True
        )
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Remove the y-axis label
        ax.set_ylabel(None)
        
        # Remove the x-axis label
        ax.set_xlabel(None)
        
        # Show the hypothesis value as a vertical line with a label
        ax.axvline(
            x=hypothesized_value,
            ymax=0.97-.02,
            color="#262626",
            linestyle="--",
            linewidth=1.5,
            alpha=0.5
        )
        ax.text(
            x=hypothesized_value, 
            y=plt.ylim()[1] * 0.97, 
            s='Hypothesized value: {:.2f}'.format(hypothesized_value)+' ({:.1%})'.format(p_value),
            horizontalalignment='center',
            fontname="Arial",
            fontsize=9,
            color="#262626",
            alpha=0.75
        )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()

    # Print interpretation of results
    if alternative_hypothesis == "two-sided" and ttest_results.pvalue <= (1-confidence_interval):
        print("There is a statistically detectable difference between the hypothesized mean and the sample mean.")
    elif ttest_results.pvalue <= (1-confidence_interval):
        print("The sample mean is " + alternative_hypothesis + " than the hypothesized mean to a statistically detectable extent.")
    else:
        print("Cannot reject the null hypothesis.")
    print("p-value:", str(round(ttest_results.pvalue, 3)))
    print("statistic:", str(round(ttest_results.statistic, 3)))
    
    # Return results
    return ttest_results




# Load packages
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

# Declare function
def PlotVectors(list_of_vectors,
                list_of_vector_labels=None,
                color_palette='Set2'):
    """
    Plots a list of 2D vectors.

    Args:
        list_of_vectors (list): A list of 2D vectors represented as lists or numpy arrays.
        list_of_vector_labels (list, optional): A list of labels for the vectors. If None, default labels will be used. Defaults to None.
        color_palette (str, optional): The name of the seaborn color palette to use. Defaults to 'Set2'.

    Raises:
        Exception: If a vector has more than 2 coordinates.

    Returns:
        None
    """
    
    # Iterate through the list of vectors, and if the vector is a list, convert it to a numpy array
    for vector_index in range(len(list_of_vectors)):
        # If the vector is a list, convert it to a numpy array
        if type(list_of_vectors[vector_index]) == list:
            list_of_vectors[vector_index] = np.array(list_of_vectors[vector_index])
        # If the vector has more than 2 coordinates, raise an exception
        if len(list_of_vectors[vector_index].shape) > 2:
            raise Exception('Vector {} has more than 2 coordinates'.format(vector_index))
    
    # If list_of_vector_labels is None, generate a list of labels
    if list_of_vector_labels is None:
        list_of_vector_labels= []
        for vector_index in range(len(list_of_vectors)):
            list_of_vector_labels.append('Vector {}'.format(vector_index))
    
    # Set the color palette
    sns.set_palette(color_palette)
    
    # Plot the vectors
    _, ax = plt.subplots(figsize=(10, 10))
    ax.tick_params(axis='x', labelsize=14)
    ax.tick_params(axis='y', labelsize=14)
    
    # Set the x and y limits
    for vector in list_of_vectors:
        print(vector.min())
    x_min = min([0] + [vector.min() for vector in list_of_vectors])
    x_max = max([0] + [vector.max() for vector in list_of_vectors])
    y_min = min([0] + [vector.min()  for vector in list_of_vectors])
    y_max = max([0] + [vector.max()  for vector in list_of_vectors])
    ax.set_xlim([x_min-1, x_max+1])
    ax.set_ylim([y_min-1, y_max+1])
        
    # Plot each vector
    for i, v in enumerate(list_of_vectors):
        sgn = 0.4 * np.array([[1] if i==0 else [i] for i in np.sign(v)])
        plt.quiver(
            v[0], 
            v[1],
            color=sns.color_palette()[i],
            angles='xy', 
            scale_units='xy', 
            scale=1
        )
        ax.text(
            v[0]-0.2+sgn[0],
            v[1]-0.2+sgn[1], 
            list_of_vector_labels[i],
            fontsize=10,
            color=sns.color_palette()[i]
        )
    
    # Add a grid
    plt.grid()
    
    # Set aspect ratio to equal
    plt.gca().set_aspect("equal")
    
    # Remove borders
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    plt.gca().spines['bottom'].set_visible(False)
    plt.gca().spines['left'].set_visible(False)
    
    # Add line at x=0 and y=0
    plt.axhline(y=0, color='k', alpha=0.25)
    plt.axvline(x=0, color='k', alpha=0.25)
    
    # Show the plot
    plt.show()




# Load packages
import pandas as pd
import os
from math import ceil
from matplotlib import pyplot as plt
import seaborn as sns
import textwrap

# Declare function
def PlotClusteredBarChart(dataframe,
                          grouping_column_name_1, 
                          grouping_column_name_2,
                          value_column_name,
                          # Plot formatting arguments
                          color_palette="Set1",
                          fill_transparency=0.8,
                          display_order_list=None,
                          figure_size=(6, 6),
                          show_legend=True,
                          # Text formatting arguments
                          title_for_plot=None,
                          subtitle_for_plot=None,
                          caption_for_plot=None,
                          data_source_for_plot=None,
                          x_indent=-0.04,
                          title_y_indent=1.15,
                          subtitle_y_indent=1.1,
                          caption_y_indent=-0.15,
                          decimal_places_for_data_label=0):
    """
    This function creates a clustered bar chart using seaborn. 
    It takes a pandas DataFrame and column names for grouping and value, along with various optional parameters for plot formatting and text formatting. 

    Args:
        dataframe (pd.DataFrame): The data to plot.
        grouping_column_name_1 (str): The name of the first grouping column.
        grouping_column_name_2 (str): The name of the second grouping column.
        value_column_name (str): The name of the value column.
        color_palette (str, optional): The color palette to use for the plot. Defaults to "Set1".
        fill_transparency (float, optional): The transparency of the bars. Defaults to 0.8.
        display_order_list (list, optional): The order to display the 1st categories in the plot. Defaults to None.
        figure_size (tuple, optional): The size of the figure. Defaults to (6, 6).
        title_for_plot (str, optional): The title for the plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to None.
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        x_indent (float, optional): The x-indent for the title and subtitle. Defaults to -0.04.
        title_y_indent (float, optional): The y-indent for the title. Defaults to 1.15.
        subtitle_y_indent (float, optional): The y-indent for the subtitle. Defaults to 1.1.
        caption_y_indent (float, optional): The y-indent for the caption. Defaults to -0.15.
        decimal_places_for_data_label (int, optional): The number of decimal places for the data labels. Defaults to 0.
    """
    # If display_order_list is provided, check that it contains all of the categories in the dataframe
    if display_order_list != None:
        if not set(display_order_list).issubset(set(dataframe[grouping_column_name_1].unique())):
            raise ValueError("display_order_list must contain all of the categories in the " + grouping_column_name_1 + " column of dataframe.")
        else:
            # Order the dataframe by the display_order_list
            dataframe[grouping_column_name_1] = pd.Categorical(dataframe[grouping_column_name_1], categories=display_order_list, ordered=True)
    else:
        # If display_order_list is not provided, create one from the dataframe
        display_order_list = dataframe.sort_values(value_column_name, ascending=True)[grouping_column_name_1].unique()
    
    # Draw a nested barplot by species and sex
    ax = sns.catplot(
        data=dataframe, 
        kind="bar",
        x=grouping_column_name_1, 
        y=value_column_name, 
        hue=grouping_column_name_2,
        palette=color_palette, 
        alpha=fill_transparency, 
        height=figure_size[1],
        aspect=figure_size[0] / figure_size[1],
        ci=None,
    )
    
    # Remove the legend if show_legend is False
    if show_legend == False:
        ax.ax.legend_.remove()
    
    # Add space between the title and the plot
    plt.subplots_adjust(top=0.85)
    
    # Wrap x axis label using textwrap
    wrapped_variable_name = "\n".join(textwrap.wrap(grouping_column_name_1, 40, break_long_words=False))  # String wrap the variable name
    ax.ax.set_xlabel(wrapped_variable_name)
    
    # Format and wrap x axis tick labels using textwrap
    x_tick_labels = ax.ax.get_xticklabels()
    wrapped_x_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 40, break_long_words=False)) for label in x_tick_labels]
    ax.ax.set_xticklabels(
        wrapped_x_tick_labels, 
        fontsize=10, 
        # fontname="Arial", 
        color="#262626"
    )
    
    # Change x-axis colors
    ax.ax.tick_params(axis='x', colors="#262626")
    ax.ax.spines['top'].set_color("#666666")
    
    # Remove bottom, left, and right spines
    ax.ax.spines['bottom'].set_visible(False)
    ax.ax.spines['right'].set_visible(False)
    ax.ax.spines['left'].set_visible(False)
    
    # Remove the y-axis tick text
    ax.ax.set_yticklabels([])
    
    # Add data labels to each bar in the plot
    # Get minimum value absolute value of the data
    min_value = dataframe[value_column_name].abs().min()
    for p in ax.ax.patches:
        # Get the height of each bar
        height = p.get_height()
        
        # Format the height to the specified number of decimal places
        height = round(height, decimal_places_for_data_label)
        
        # Add the data label to the plot, using the specified number of decimal places
        value_label = "{:,.{decimal_places}f}".format(height, decimal_places=decimal_places_for_data_label)
        ax.ax.text(
            x=p.get_x() + p.get_width() / 2.,
            y=height + (min_value * 0.05),
            s=value_label,
            ha="center",
            # fontname="Arial",
            fontsize=10,
            color="#262626"
        )
        
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.ax.transAxes
    )
    
        
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.ax.transAxes
        )
    
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import ttest_rel
import seaborn as sns
import textwrap

# Declare function
def TwoSampleTTestPaired(dataframe,
                         outcome1_column,
                         outcome2_column,
                         confidence_interval=.95,
                         alternative_hypothesis="two-sided",
                         null_handling='omit',
                         plot_sample_distributions=True,
                         fill_color="#999999",
                         fill_transparency=0.2,
                         title_for_plot="Paired T-Test",
                         subtitle_for_plot="Shows the distribution of the mean difference across all pairs",
                         caption_for_plot=None,
                         data_source_for_plot=None,
                         show_y_axis=False,
                         title_y_indent=1.1,
                         subtitle_y_indent=1.05,
                         caption_y_indent=-0.15,
                         figure_size=(8, 6)):
    """
    Conducts a paired two-sample t-test on two columns of a pandas DataFrame.

    Args:
        dataframe (pandas.DataFrame): The DataFrame containing the data to be analyzed.
        outcome1_column (str): The name of the first column containing the outcome data.
        outcome2_column (str): The name of the second column containing the outcome data.
        confidence_interval (float, optional): The desired level of confidence for the test. Defaults to 0.95.
        alternative_hypothesis (str, optional): The alternative hypothesis to be tested. Can be "two-sided", "greater", or "less". Defaults to "two-sided".
        null_handling (str, optional): How to handle null values in the data. Can be "omit", "raise", or "propagate". Defaults to "omit".
        plot_sample_distributions (bool, optional): Whether to plot the distribution of the mean difference across all pairs. Defaults to True.
        fill_color (str, optional): The color to fill the distribution plot with. Defaults to "#999999".
        fill_transparency (float, optional): The transparency of the fill color. Defaults to 0.2.
        title_for_plot (str, optional): The title of the distribution plot. Defaults to "Paired T-Test".
        subtitle_for_plot (str, optional): The subtitle of the distribution plot. Defaults to "Shows the distribution of the mean difference across all pairs".
        caption_for_plot (str, optional): The caption of the distribution plot. Defaults to None.
        data_source_for_plot (str, optional): The data source of the distribution plot. Defaults to None.
        show_y_axis (bool, optional): Whether to show the y-axis on the distribution plot. Defaults to False.
        title_y_indent (float, optional): The y-indent of the title on the distribution plot. Defaults to 1.1.
        subtitle_y_indent (float, optional): The y-indent of the subtitle on the distribution plot. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent of the caption on the distribution plot. Defaults to -0.15.
        figure_size (tuple, optional): The size of the distribution plot. Defaults to (8, 6).

    Returns:
        scipy.stats.Ttest_relResult: The results of the paired two-sample t-test.
    """
    
    # Show pivot table summary
    table_groups = dataframe.groupby(
        [True]*len(dataframe)
    ).agg(
        {
            outcome1_column: [min, max, np.mean, np.std],
            outcome2_column: [min, max, np.mean, np.std]
        }
    )
    print(table_groups)
    print("\n")
    
    # Conduct two sample t-test
    ttest_results = ttest_rel(
        dataframe[outcome1_column], 
        dataframe[outcome2_column], 
        axis=0,
        nan_policy=null_handling, 
        alternative=alternative_hypothesis
    )
    
    # Get the p-value
    p_value = ttest_results.pvalue
    
    # Get the test statistic
    test_statistic = ttest_results.statistic
    
    if plot_sample_distributions:
        # Calculate the differences in each observation
        dataframe['Difference'] = dataframe[outcome1_column] - dataframe[outcome2_column]
        
        # Calculate the average difference
        sample_mean = dataframe['Difference'].mean()
        
        # Calculate the standard error of the mean
        standard_error = dataframe['Difference'].sem()
        
        # Generate normal distribution of differences
        dataframe = pd.DataFrame(
            np.random.normal(loc=sample_mean,
                            scale=standard_error,
                            size=10000),
            columns=['Difference']
        )
            
        # Plot distribution of means
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
        
        # Create histogram using seaborn
        ax = sns.kdeplot(
            data=dataframe,
            x='Difference',
            color=fill_color,
            alpha=fill_transparency,
            fill=True,
            common_norm=True
        )
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Remove the y-axis label
        ax.set_ylabel(None)
        
        # Remove the x-axis label
        ax.set_xlabel(None)
        
        # Show the hypothesis value as a vertical line with a label
        ax.axvline(
            x=0,
            ymax=0.97-.02,
            color="#262626",
            linestyle="--",
            linewidth=1.5,
            alpha=0.5
        )
        ax.text(
            x=0, 
            y=plt.ylim()[1] * 0.97, 
            s='No difference ({:.1%})'.format(p_value),
            horizontalalignment='center',
            fontname="Arial",
            fontsize=9,
            color="#262626",
            alpha=0.75
        )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()

    # Print interpretation of results
    if alternative_hypothesis == "two-sided" and ttest_results.pvalue <= (1-confidence_interval):
        print("There is a statistically detectable difference between the sample means of " + outcome1_column + " and " + outcome2_column + ".")
    elif ttest_results.pvalue <= (1-confidence_interval):
        print("The sample mean of " + outcome1_column + " is " + alternative_hypothesis + " than that of " + outcome2_column + " to a statistically detectable extent.")
    else:
        print("Cannot reject the null hypothesis.")
    print("p-value:", str(round(p_value, 3)))
    print("statistic:", str(round(test_statistic, 3)))
    
    # Return results
    return ttest_results




from .CreateMetalogDistributionFromPercentiles import CreateMetalogDistributionFromPercentiles
from .CreateMetalogDistribution import CreateMetalogDistribution
from .CreateSIPDataframe import CreateSIPDataframe
from .SimulateCountOfSuccesses import SimulateCountOfSuccesses
from .SimulateCountOutcome import SimulateCountOutcome
from .SimulateCountUntilFirstSuccess import SimulateCountUntilFirstSuccess
from .SimulateNormallyDistributedOutcome import SimulateNormallyDistributedOutcome
from .SimulateTDistributedOutcome import SimulateTDistributedOutcome
from .SimulateTimeBetweenEvents import SimulateTimeBetweenEvents
from .SimulateTimeUntilNEvents import SimulateTimeUntilNEvents


# Import packages
import numpy as np
import sympy as sp

# Declare function
def ConvertMatrixToRowEchelonForm(matrix,
                                  show_pivot_columns=False):
    """
    Converts a matrix to row echelon form.
    
    Args:
        matrix (np.array or list): Matrix to be converted to row echelon form.
        show_pivot_columns (bool, optional): If True, the function will print the pivot columns. Defaults to False.

    Returns:
        np.array: Matrix in row echelon form.
    """
    
    # If matrix is list, convert it to a matrix
    if type(matrix) == list:
        matrix = np.array(matrix)
    
    # Convert matrix to row echelon form
    matrix_row_echelon = sp.Matrix(matrix).rref()
    
    # Show rank of matrix
    print("Rank of matrix: " + str(len(matrix_row_echelon[1])))
    
    # Print list of pivot columns, if requested
    if show_pivot_columns:
        print("Pivot columns: " + str(matrix_row_echelon[1]))
    
    # Convert back to numpy array
    matrix_row_echelon = np.array(matrix_row_echelon[0])
    
    # Return matrix in row echelon form
    return matrix_row_echelon




# Load packages
import matplotlib.pyplot as plt
import numpy as np

# Declare functions
def CalculateEigenvalues(matrix, 
                         show_plot=True, 
                         matrix_color='#033dfc', 
                         eigenvector_color='#e82a53', 
                         x_min=-1, 
                         x_max=5, 
                         y_min=-1, 
                         y_max=5, 
                         show_labels=True, 
                         plot_with_grid=True):
    """
    Calculates the eigenvalues and eigenvectors of a 2x2 matrix and optionally plots the matrix and eigenvectors.
    
    Args:
        matrix (numpy.ndarray or list): The 2x2 matrix to calculate the eigenvalues and eigenvectors of.
        show_plot (bool, optional): Whether to show a plot of the matrix and eigenvectors. Defaults to True.
        matrix_color (str, optional): The color to use for the matrix plot. Defaults to '#033dfc'.
        eigenvector_color (str, optional): The color to use for the eigenvector plot. Defaults to '#e82a53'.
        x_min (int, optional): The minimum x value for the plot. Defaults to -1.
        x_max (int, optional): The maximum x value for the plot. Defaults to 5.
        y_min (int, optional): The minimum y value for the plot. Defaults to -1.
        y_max (int, optional): The maximum y value for the plot. Defaults to 5.
        show_labels (bool, optional): Whether to show labels on the plot. Defaults to True.
        plot_with_grid (bool, optional): Whether to show a grid on the plot. Defaults to True.
    
    Returns:
        None
    
    Raises:
        ValueError: If the matrix is not 2x2.
    
    """
    
    # If matrix is list, convert to numpy array
    if type(matrix) == list:
        matrix = np.array(matrix)
    
    # Calculate eigenvalues and eigenvectors of a matrix
    eigenvalues, eigenvectors = np.linalg.eig(matrix)

    # If the plot is to be shown, plot the eigenvectors and the matrix
    if show_plot:
        # Ensure that matrix is 2x2
        if matrix.shape != (2, 2):
            print('Matrix must be 2x2 to plot')
        else:
            # Perform matrix multiplication
            transformed_square = np.dot(matrix, np.array([[0,1,1,0,0],[0,0,1,1,0]]))

            # Plot the vectors
            _, ax = plt.subplots(figsize=(10, 10))
            ax.tick_params(axis='x', labelsize=14)
            ax.tick_params(axis='y', labelsize=14)
            
            # Set the x and y limits
            ax.set_xlim([x_min-1, x_max+1])
            ax.set_ylim([y_min-1, y_max+1])

            # Add a grid, if requested
            if plot_with_grid:
                plt.grid()
            
            # Set aspect ratio to equal
            plt.gca().set_aspect("equal")
            
            # Remove borders
            plt.gca().spines['top'].set_visible(False)
            plt.gca().spines['right'].set_visible(False)
            plt.gca().spines['bottom'].set_visible(False)
            plt.gca().spines['left'].set_visible(False)
            
            # Add line at x=0 and y=0
            plt.axhline(y=0, color='k', alpha=0.25)
            plt.axvline(x=0, color='k', alpha=0.25)
            
            # Plot the unit square
            plt.plot([0,1,1,0,0],[0,0,1,1,0],
                    color='black',
                    alpha=0.5)
            plt.fill([0,1,1,0,0],[0,0,1,1,0],
                    color='black',
                    alpha=0.1)
            # Add label for unit square
            if show_labels:
                plt.text(
                    0.5, 
                    -0.15, 
                    "Unit Square", 
                    fontsize=8, 
                    fontfamily='Arial', 
                    ha='center', 
                    va='center'
                )
            
            # Plot the transformed unit square
            plt.plot(
                transformed_square[0,:],
                transformed_square[1,:],
                color=matrix_color,
                alpha=0.25
            )
            plt.fill(
                transformed_square[0,:],
                transformed_square[1,:],
                color=matrix_color,
                alpha=0.1
            )
            # Add label for transformed unit square
            if show_labels:
                plt.text(
                    transformed_square[0,2]+0.15,
                    transformed_square[1,2]+0.15, 
                    "Transformed Unit Square #1", 
                    fontsize=8, 
                    fontfamily='Arial', 
                    ha='center', 
                    va='center', 
                    color=matrix_color
                )
            
            # If eigenvalues are complex, print a warning
            if ~np.iscomplex(eigenvalues).any():
                # Plot the eigenvector
                plt.quiver(
                    [0, 0],
                    [0, 0],
                    [eigenvectors[0], eigenvectors[0]], 
                    [eigenvectors[1], eigenvectors[1]], 
                    color=eigenvector_color, 
                    angles='xy', 
                    scale_units='xy', 
                    scale=1
                )
                
            # Show the plot
            plt.show()
    
    # Print the eigenvalues
    print('Eigenvalues: {}'.format(eigenvalues))
    
    # Print the eigenvectors
    print('Eigenvectors: {}'.format(eigenvectors))
    
    # If matrix is an identity matrix, print a warning
    if np.array_equal(matrix, np.identity(2)):
        print('Warning: Matrix is an identity matrix, meaning any vector is an eigenvector.')
    
    # If eigenvalues are complex, print a warning
    if np.iscomplex(eigenvalues).any():
        print('Warning: Eigenvalues are complex, meaning there are no real eigenvectors.')
    
    # If eignvectors are equivalent to identity matrix, print a warning
    if np.array_equal(eigenvectors, np.identity(len(eigenvalues))):
        print('Warning: Eigenvectors are equivalent to identity matrix, meaning any vector is an eigenvector.')
    
    # Calculate the determinant of the coefficient matrix
    determinant = np.linalg.det(eigenvectors)
    if determinant == 0:
        print("The system of eigenvector equations is singular, and does not have a unique solution. At least two equations are linearly dependent.")
    else:
        print("The system of eigenvector equations is non-singular, and has a unique solution. The equations are linearly independent.")




# Load packages
from lifelines import CoxPHFitter
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.api as sm
import textwrap

# Declare function
def ConductCoxProportionalHazardRegression(dataframe,
                                           outcome_column,
                                           duration_column,
                                           list_of_predictor_columns,
                                           plot_survival_curve=True,
                                           line_color="#3269a8",
                                           line_alpha=0.8,
                                           color_palette="Set2",
                                           markers="o",
                                           number_of_x_axis_ticks=20,
                                           x_axis_tick_rotation=None,
                                           title_for_plot="Survival Counts by Group",
                                           subtitle_for_plot='Shows the counts of "surviving" records by group over time',
                                           caption_for_plot=None,
                                           data_source_for_plot=None,
                                           x_indent=-0.127,
                                           title_y_indent=1.125,
                                           subtitle_y_indent=1.05,
                                           caption_y_indent=-0.3,
                                           figure_size=(8, 5)):
    """
    Conducts Cox Proportional Hazard Regression on the given dataframe and returns the model.
    
    Args:
        dataframe (pandas.DataFrame): The dataframe containing the data to be used in the regression.
        outcome_column (str): The name of the column containing the outcome variable.
        duration_column (str): The name of the column containing the duration variable.
        list_of_predictor_columns (list): A list of column names containing the predictor variables.
        plot_survival_curve (bool, optional): Whether or not to plot the survival curve. Defaults to True.
        line_color (str, optional): The color of the lines in the plot. Defaults to "#3269a8".
        line_alpha (float, optional): The alpha value of the lines in the plot. Defaults to 0.8.
        color_palette (str, optional): The color palette to use in the plot. Defaults to "Set2".
        markers (str, optional): The marker style to use in the plot. Defaults to "o".
        number_of_x_axis_ticks (int, optional): The number of ticks to use on the x-axis. Defaults to 20.
        x_axis_tick_rotation (int, optional): The rotation angle of the x-axis tick labels. Defaults to None.
        title_for_plot (str, optional): The title of the plot. Defaults to "Survival Counts by Group".
        subtitle_for_plot (str, optional): The subtitle of the plot. Defaults to 'Shows the counts of "surviving" records by group over time'.
        caption_for_plot (str, optional): The caption of the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source of the plot. Defaults to None.
        x_indent (float, optional): The x-axis indent of the plot. Defaults to -0.127.
        title_y_indent (float, optional): The title y-axis indent of the plot. Defaults to 1.125.
        subtitle_y_indent (float, optional): The subtitle y-axis indent of the plot. Defaults to 1.05.
        caption_y_indent (float, optional): The caption y-axis indent of the plot. Defaults to -0.3.
        figure_size (tuple, optional): The size of the plot. Defaults to (8, 5).
    
    Returns:
        lifelines.CoxPHFitter: The Cox Proportional Hazard Regression model.
    """
    
    # Select the columns to keep
    dataframe = dataframe[[outcome_column, duration_column] + list_of_predictor_columns].copy()

    # Remove records with missing values
    dataframe = dataframe.dropna()

    # Order the dataframe by outcome
    dataframe = dataframe.sort_values(by = [outcome_column, duration_column], ascending = [True, True])

    # Create a new column showing unqiue status as group starting with 0
    unique_status = dataframe[outcome_column].unique()
    dataframe["Survival Group"] = dataframe[outcome_column].apply(lambda x: list(unique_status).index(x))
    # Print the survival group and their corresponding status
    print("Outcome variables have been assigned accordingly to each group:")
    print(dataframe[["Survival Group", outcome_column]].drop_duplicates().reset_index(drop = True))
    # Drop the survival group column
    dataframe = dataframe.drop(columns = [outcome_column])
    
    # Create a time series dataframe using the duration column
    # Get the maximum duration
    max_duration = dataframe[duration_column].max()
    # Create a time series dataframe
    time_series_dataframe = pd.DataFrame(
        data = range(0, max_duration + 1),
        columns = [duration_column]
    )
    
    # Iterate through the status groups
    for status_group in dataframe["Survival Group"].unique():
        # Create temporary copy of time series dataframe
        temp_time_series_dataframe = time_series_dataframe.copy()
        
        # Add the status group to the temporary time series dataframe
        temp_time_series_dataframe["Survival Group"] = status_group
        
        # Filter the original dataframe to the status group
        df_temp = dataframe[dataframe["Survival Group"] == status_group].copy()
        
        # Get the number of records in the dataframe that have a duration less than or equal to the time series dataframe
        temp_time_series_dataframe["Survival Count"] = temp_time_series_dataframe[duration_column].apply(
            lambda x: len(df_temp[df_temp[duration_column] > x])
        )
        
        # Row bind the temporary time series dataframe to the time series dataframe
        time_series_dataframe = pd.concat([time_series_dataframe, temp_time_series_dataframe], axis=0)
    # Filter out records with no survival group
    time_series_dataframe = time_series_dataframe[time_series_dataframe["Survival Group"].notnull()].reset_index(drop=True)
    
    # Use the time series dataframe to plot the survival curve
    if plot_survival_curve:
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
    
        # Use Seaborn to create a line plot
        sns.lineplot(
            data=time_series_dataframe,
            x=duration_column,
            y="Survival Count",
            hue="Survival Group",
            palette=color_palette,
            alpha=line_alpha,
            marker=markers,
            ax=ax
        )
        
        # Remove top and right spines, and set bottom and left spines to gray
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color('#666666')
        ax.spines['left'].set_color('#666666')
    
        # Set the number of ticks on the x-axis
        if number_of_x_axis_ticks is not None:
            ax.xaxis.set_major_locator(plt.MaxNLocator(number_of_x_axis_ticks))
            
        # Rotate x-axis tick labels
        if x_axis_tick_rotation is not None:
            plt.xticks(rotation=x_axis_tick_rotation)

        # Format tick labels to be Arial, size 9, and color #666666
        ax.tick_params(
            which='major',
            labelsize=9,
            color='#666666'
        )
        
        # Format tick labels to be Arial, size 9, and color #666666
        ax.tick_params(
            which='major',
            labelsize=9,
            color='#666666'
        )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Word wrap the subtitle without splitting words
        if subtitle_for_plot != None:   
            subtitle_for_plot = textwrap.fill(subtitle_for_plot, 100, break_long_words=False)
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Move the y-axis label to the top of the y-axis, and set the font to Arial, size 9, and color #666666
        ax.yaxis.set_label_coords(-0.1, 0.84)
        ax.yaxis.set_label_text(
            "Survival Count",
            fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
        # Move the x-axis label to the right of the x-axis, and set the font to Arial, size 9, and color #666666
        ax.xaxis.set_label_coords(0.9, -0.1)
        ax.xaxis.set_label_text(
            duration_column,
            fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )

        # Show plot
        plt.show()
    
    
    # Create Cox Proportional hazard regression model
    model = CoxPHFitter()
    model.fit(
        df=dataframe, 
        duration_col=duration_column, 
        event_col="Survival Group"
    )
    model.print_summary()

    # Return the model
    return model




# Load packages
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

# Declare function
def FindMinimumSquareLoss(observed_values,
                          predicted_values,
                          show_plot=True):
    """
    This function calculates the minimum square loss between observed and predicted values.

    Args:
        observed_values: A list of observed values.
        predicted_values: A list of predicted values.

    Returns:
        The minimum square loss between the observed and predicted values.
    """
    # Check if the number of observed values is equal to the number of predicted values.
    if len(observed_values) != len(predicted_values):
        raise ValueError("The number of observed values must be equal to the number of predicted values.")
    
    # If values are passed as a list, convert them to a numpy array.
    if type(observed_values) == list:
        observed_values = np.array(observed_values)
    if type(predicted_values) == list:
        predicted_values = np.array(predicted_values)    
    
    # Count the number of observed values.
    n = len(observed_values)
    
    # Calculate the prediction error.
    sq_prediction_error = (predicted_values - observed_values) ** 2
    
    # Calculate the sum of squares.
    sum_of_squares = sum(sq_prediction_error)
    
    # Calculate the minimum square loss.
    FindMinimumSquareLoss = sum_of_squares / n
    
    # Plot the prediction errors in a plot, if requested.
    if show_plot:
        sns.boxplot(y=sq_prediction_error, color="grey")
        # Add the minimum square loss as a horizontal line.
        plt.axhline(FindMinimumSquareLoss, color="r", linestyle="--")
        # Add labels.
        plt.title("Squared Prediction Errors", fontsize=14, fontweight="bold")
        plt.text(0.5, FindMinimumSquareLoss + .01, "Min. Sq. Loss = " + str(round(FindMinimumSquareLoss, 2)))
        # Remove borders from plot
        sns.despine(bottom=True)
        # Show plot
        plt.show()
    
    # Return the minimum square loss.
    return FindMinimumSquareLoss




# Load packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.manifold import TSNE

# Delcare function
def ConductManifoldLearning(dataframe,
                            list_of_numeric_columns=None,
                            number_of_components=3,
                            random_seed=412,
                            show_component_summary_plots=True,
                            summary_plot_size=(20, 20)):
    """
    Conducts manifold learning on a given dataframe and returns a new dataframe with the
    original columns and the new manifold learning components.

    Args:
        dataframe (pandas.DataFrame): The input dataframe.
        list_of_numeric_columns (list, optional): A list of column names to use for the manifold
            learning. If None, all numeric columns will be used. Defaults to None.
        number_of_components (int, optional): The number of components to generate. Defaults to 3.
        random_seed (int, optional): The random seed to use for the manifold learning algorithm.
            Defaults to 412.
        show_component_summary_plots (bool, optional): Whether to show summary plots of each
            component for each variable. Defaults to True.
        summary_plot_size (tuple, optional): The size of the summary plots. Defaults to (20, 20).

    Returns:
        pandas.DataFrame: A new dataframe with the original columns and the new manifold learning
        components.
    """
    
    # If list_of_numeric_columns is not specified, then use all numeric variables
    if list_of_numeric_columns is None:
        list_of_numeric_columns = dataframe.select_dtypes(include=[np.number]).columns.tolist()
        
    # Select only numeric variables
    dataframe_manifold = dataframe[list_of_numeric_columns].copy()
    
    # Remove missing values
    dataframe_manifold = dataframe_manifold.dropna()

    # Create list of column names
    list_of_column_names = dataframe_manifold.columns.tolist()

    # Conduct Manifold Learning
    model = TSNE(
        n_components=number_of_components,
        random_state=random_seed
    )
    components = model.fit_transform(dataframe_manifold)
    
    # Get column names for new components
    list_of_component_names = []
    for i in range(1, number_of_components + 1):
        list_of_component_names.append("MLC" + str(i))
    
    # Change component column names
    components = pd.DataFrame(
        data=components,
        columns=list_of_component_names
    )

    # Add Manifold Learning components to original dataframe
    dataframe = pd.concat([dataframe, components], axis=1)
    
    # If requested, show box plots of each component for each variable
    # Put each numeric variables on the Y axis and each component on the X axis
    if show_component_summary_plots:
        plt.figure(figsize=summary_plot_size)
        sns.pairplot(
            data=dataframe[list_of_numeric_columns + list_of_component_names],
            x_vars=list_of_component_names,
            y_vars=list_of_numeric_columns,
            kind='kde'
        )
        plt.suptitle("Component Summary Plots", fontsize=15)
        plt.show()
    
    # Return dataframe
    return(dataframe)




# Load packages
import pandas as pd
import os

# Declare function
def ImportDataFromFolder(folder_path,
                         force_column_names_to_match = True):
    """
    Imports all CSV and Excel files from a folder and combines them into a single dataframe.
    Note: Each file should have the same column names.

    Args:
        folder_path (str): The path to the folder containing the csv and Excel files.
        force_column_names_to_match (bool, optional): Whether to force the column names to match across all files. Defaults to True.

    Returns:
        Pandas dataframe: A dataframe containing all the data from the csv and Excel files.
    """
    
    # Create empty list to store dataframes
    df_list = []
    
    # Loop through all files in the folder
    for file in os.listdir(folder_path):
        # Check if file is a csv file
        if file.endswith(".csv"):
            # Import data from csv file
            df = pd.read_csv(folder_path + "/" + file)
            # Append dataframe to list
            df_list.append(df)
        # Check if file is an excel file
        elif file.endswith(".xlsx"):
            # Import data from excel file
            df = pd.read_excel(folder_path + "/" + file)
            # Append dataframe to list
            df_list.append(df)
    
    # Ensure that column names match across all dataframes
    for i in range(0, len(df_list)):
        # Check if column names match
        if list(df_list[i]) != list(df_list[0]):
            if force_column_names_to_match:
                # Rename columns
                df_list[i].columns = list(df_list[0])
            else:
                # Throw error
                raise ValueError("Column names do not match across all dataframes.")
    
    # Row bind all dataframes
    df_list = pd.concat(df_list, axis = 0)
    
    # Return list of dataframes
    return(df_list)




# Load packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn import metrics
from sklearn.model_selection import train_test_split
import tensorflow as tf

# Declare function
def CreateNeuralNetwork_SingleOutcome(dataframe,
                                      outcome_variable,
                                      list_of_predictor_variables,
                                      number_of_hidden_layers,
                                      is_outcome_categorical=True,
                                      test_size=0.2,
                                      scale_predictor_variables=True,
                                      plot_loss=True,
                                      plot_model_test_performance=True,
                                      show_predictor_ranges=True,
                                      initial_learning_rate=0.01,
                                      number_of_steps_gradient_descent=100,
                                      lambda_for_regularization=0.001,
                                      random_seed=412):
    # Keep only the predictors and outcome variable
    dataframe = dataframe[list_of_predictor_variables + [outcome_variable]].copy()
    
    # Set the random seed
    tf.random.set_seed(random_seed)
    
    # Use Keras to create a normalization layer
    if scale_predictor_variables:
        normalizer = tf.keras.layers.Normalization(axis=-1)
        normalizer.adapt(np.array(dataframe[list_of_predictor_variables]))  # Learns the statistics of the data
        norm_predictor_variables = normalizer(np.array(dataframe[list_of_predictor_variables]))
        dataframe[list_of_predictor_variables] = pd.DataFrame(norm_predictor_variables.numpy(), columns=list_of_predictor_variables)

    # Show the peak-to-peak range of each predictor
    if show_predictor_ranges:
        print("\nPeak-to-peak range of each predictor:")
        print(np.ptp(dataframe[list_of_predictor_variables], axis=0))
        
    # Split dataframe into training and test sets
    train, test = train_test_split(
        dataframe, 
        test_size=test_size,
        random_state=random_seed
    )
    
    # Choose activation function
    if is_outcome_categorical: 
        if len(dataframe[outcome_variable].unique()) == 2:
            activation_function = 'sigmoid'
        else:
            activation_function = 'linear'
    else:
        if dataframe[outcome_variable].min() >= 0:
            activation_function = 'relu'
        else:
            activation_function = 'linear'
                
    # Create dictionary of layers to be used in the neural network
    dict_layers = {}
    
    # Create input layer
    dict_layers['Input layer'] = tf.keras.layers.Input(
        shape=(len(list_of_predictor_variables),), 
        name='input_layer'
    )
    
    # Create hidden layers
    for i in range(number_of_hidden_layers):
        key_text = 'Hidden layer ' + str(i + 1)
        if i < 10:
            layer_name = 'layer_0' + str(i + 1)
        else:
            layer_name = 'layer_' + str(i + 1)
        dict_layers[key_text] = tf.keras.layers.Dense(
            10 + ((number_of_hidden_layers - 1 - i) * 10), 
            activation='relu', 
            name=layer_name,
            kernel_regularizer=tf.keras.regularizers.l2(lambda_for_regularization)
        )
        
    # Create output layer
    if is_outcome_categorical and len(dataframe[outcome_variable].unique()) > 2:
        dict_layers['Output layer'] = tf.keras.layers.Dense(
            len(dataframe[outcome_variable].unique()), 
            activation=activation_function, 
            name='softmax_layer',
            kernel_regularizer=tf.keras.regularizers.l2(lambda_for_regularization)
        )
    else:
        dict_layers['Output layer'] = tf.keras.layers.Dense(
            1, 
            activation=activation_function, 
            name='final_layer',
            kernel_regularizer=tf.keras.regularizers.l2(lambda_for_regularization)
        )
    
    # Create list of layers to be used in the neural network
    list_of_layers = [dict_layers['Input layer']]
    for i in range(number_of_hidden_layers):
        key_text = 'Hidden layer ' + str(i + 1)
        list_of_layers.append(dict_layers[key_text])
    list_of_layers.append(dict_layers['Output layer'])
    
    # Create sequential model
    model = tf.keras.Sequential(list_of_layers) 
    
    # Show model summary
    model.summary()
    
    # Define loss function and optimizer
    if is_outcome_categorical:
        if len(dataframe[outcome_variable].unique()) == 2:
            model.compile(
                loss=tf.keras.losses.BinaryCrossentropy(), 
                optimizer=tf.keras.optimizers.Adam(learning_rate=initial_learning_rate),
                metrics=['accuracy']
            )
        else:
            model.compile(
                loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True), 
                optimizer=tf.keras.optimizers.Adam(learning_rate=initial_learning_rate),
                metrics=['accuracy']
            )
    else:
        model.compile(
            loss=tf.keras.losses.MeanSquaredError(),
            optimizer=tf.keras.optimizers.Adam(learning_rate=initial_learning_rate),
            metrics=['mean_squared_error']
        )
    
    # Train the model
    loss_history = model.fit(
        train[list_of_predictor_variables].values,
        train[outcome_variable].values,
        epochs=number_of_steps_gradient_descent,
        verbose=0
    )
    if plot_loss:
        plt.figure(figsize=(9,9))
        sns.lineplot(
            x=range(1, number_of_steps_gradient_descent + 1),
            y=loss_history.history['loss']
        )
        plt.title('Loss Curve', size = 15)
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.show()
    
    # Test the model
    predictions = model.predict(test[list_of_predictor_variables].values)
    if is_outcome_categorical and len(dataframe[outcome_variable].unique()) > 2:
        predictions = tf.nn.softmax(predictions).numpy()  # Convert to probabilities
        predictions = pd.DataFrame(predictions).idxmax(axis=1).values  # Convert to class labels
    test['Predicted'] = predictions
    
    # Plot results of model test
    if plot_model_test_performance:
        plt.figure(figsize=(9,9))
        if is_outcome_categorical:
            if len(dataframe[outcome_variable].unique()) == 2:
                sns.scatterplot(
                    data=test,
                    x=outcome_variable,
                    y='Predicted'
                )
                plt.title('Probability vs. Observed Outcome', size = 15)
                plt.ylabel('Predicted probability')
                plt.ylim(0, 1.05)
            else:
                confusion_matrix = metrics.confusion_matrix(
                    test[outcome_variable], 
                    test['Predicted']
                )
                sns.heatmap(
                    confusion_matrix, 
                    annot=True, 
                    fmt=".3f", 
                    linewidths=.5, 
                    square=True, 
                    cmap='Blues_r'
                )
                plt.title('Confusion Matrix', size = 15)
                plt.ylabel('Actual label')
                plt.xlabel('Predicted label')
        else:
            sns.regplot(
                data=test,
                x=outcome_variable,
                y='Predicted'
            )
            plt.plot(test[outcome_variable], test[outcome_variable], color='black', alpha=0.35)
            plt.title('Predicted vs. Observed Outcome', size = 15)
        plt.show()
    
    # Return the model
    if scale_predictor_variables:
        dict_return = {
            'scaler': normalizer,
            'model': model
        }
        return(dict_return)
    else:
        return(model)




# Load packages
import numpy as np

# Declare function
def ConvertSystemOfEquationsToMatrix(coefficients,
                                     constants,
                                     show_determinant=True):
    """
    Converts a system of equations to a matrix.
    
    Args:
        coefficients (list or np.array): The coefficients of the system of equations. 
        constants (list or np.array): The constants of the system of equations.

    Returns:
        np.array: The system of equations as a matrix.
    """
    
    # If the coefficients are a list of lists, convert them to a numpy array
    if type(coefficients) == list:
        coefficients = np.array(coefficients)
        
    # Calculate the determinant of the coefficient matrix
    if show_determinant:
        determinant = np.linalg.det(coefficients)
        print("Determinant:", '{:f}'.format(determinant))
        if determinant == 0:
            print("The system of equations is singular, and does not have a unique solution. At least two equations are linearly dependent.")
        else:
            print("The system of equations is non-singular, and has a unique solution. The equations are linearly independent.")
    
    # If the constants are a list, convert them to a numpy array
    if type(constants) == list:
        constants = np.array(constants)
    
    # Add the constants to the coefficients
    matrix = np.hstack((coefficients, constants.reshape(-1, 1)))
    
    # Return the matrix
    return matrix




from .ChiSquareTestOfIndependence import ChiSquareTestOfIndependence
from .ChiSquareTestOfIndependenceFromTable import ChiSquareTestOfIndependenceFromTable
from .ConductCoxProportionalHazardRegression import ConductCoxProportionalHazardRegression
from .ConductLinearRegressionAnalysis import ConductLinearRegressionAnalysis
from .ConductLogisticRegressionAnalysis import ConductLogisticRegressionAnalysis
from .ConductSurvivalAnalysis import ConductSurvivalAnalysis
from .OneSampleTTest import OneSampleTTest
from .OneWayANOVA import OneWayANOVA
from .TTestOfMeanFromStats import TTestOfMeanFromStats
from .TTestOfProportionFromStats import TTestOfProportionFromStats
from .TTestOfTwoMeansFromStats import TTestOfTwoMeansFromStats
from .TwoSampleTTestOfIndependence import TwoSampleTTestOfIndependence
from .TwoSampleTTestPaired import TwoSampleTTestPaired


# Load packages
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def PlotOverlappingAreaChart(dataframe,
                             value_column_name,
                             variable_column_name,
                             time_column_name,
                             # Line formatting arguments
                             line_alpha=0.9,
                             color_palette="Paired",
                             # Area formatting arguments
                             fill_alpha=0.8,
                             label_variable_on_area=True,
                             label_font_size=12,
                             label_font_color="#FFFFFF",
                             # Text formatting arguments
                             number_of_x_axis_ticks=None,
                             x_axis_tick_rotation=None,
                             title_for_plot=None,
                             subtitle_for_plot=None,
                             caption_for_plot=None,
                             data_source_for_plot=None,
                             x_indent=-0.127,
                             title_y_indent=1.125,
                             subtitle_y_indent=1.05,
                             caption_y_indent=-0.3,
                             # Plot formatting arguments
                             figure_size=(8, 5)):
    """
    Plots an overlapping area chart for a time series from a given dataframe.

    Args:
        dataframe (pandas.DataFrame): The dataframe containing the data to be plotted.
        value_column_name (str): The name of the column in the dataframe that contains the values to be plotted.
        variable_column_name (str): The name of the column in the dataframe to group data by.
        time_column_name (str): The name of the column in the dataframe that contains the time data.
        line_alpha (float, optional): The transparency of the line in the plot. Defaults to 0.9.
        color_palette (str, optional): The color palette to use for the plot. Defaults to "Paired".
        fill_alpha (float, optional): The transparency of the area under the line plot. Defaults to 0.8.
        label_variable_on_area (bool, optional): Whether to label the variables on the area plot. Defaults to True.
        label_font_size (int, optional): The font size of the variable labels. Defaults to 12.
        label_font_color (str, optional): The font color of the variable labels. Defaults to "#FFFFFF".
        number_of_x_axis_ticks (int, optional): The number of ticks on the x-axis. Defaults to None.
        x_axis_tick_rotation (int, optional): The rotation of the x-axis ticks. Defaults to None.
        title_for_plot (str, optional): The title for the plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to None.
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        x_indent (float, optional): The x-indent for the plot text. Defaults to -0.127.
        title_y_indent (float, optional): The y-indent for the plot title. Defaults to 1.125.
        subtitle_y_indent (float, optional): The y-indent for the plot subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent for the plot caption. Defaults to -0.3.
        figure_size (tuple, optional): The size of the figure for the plot. Defaults to (8, 5).
    """
    
    # Create figure and axes
    fig, ax = plt.subplots(figsize=figure_size)
    
    # Order the dataframe by the mean of the value column for each variable
    averages = dataframe.groupby(variable_column_name).mean().reset_index()
    averages = averages.sort_values(by=value_column_name, ascending=False)
    dataframe[variable_column_name] = pd.Categorical(dataframe[variable_column_name], categories=averages[variable_column_name], ordered=True)
    dataframe = dataframe.sort_values(by=[variable_column_name])
    
    # Use Seaborn to create a line plot
    sns.lineplot(
        data=dataframe,
        x=time_column_name,
        y=value_column_name,
        hue=variable_column_name,
        palette=color_palette,
        alpha=line_alpha,
        ax=ax,
        legend=None if label_variable_on_area else "brief" # If label_variable_on_area is False, include a legend
    )
    
    # Create sns color palette
    sns_color_palette = sns.color_palette(color_palette)
    
    # Fill the area under the line plot for each variable
    for variable in dataframe[variable_column_name].unique():
        # Get the data for the variable
        variable_data = dataframe[dataframe[variable_column_name] == variable]
        
        # Fill the area under the line plot for the variable
        ax.fill_between(
            x=variable_data[time_column_name],
            y1=variable_data[value_column_name],
            y2=0,
            alpha=fill_alpha,
            color=sns_color_palette[dataframe[variable_column_name].unique().tolist().index(variable)] # Get the color of the variable from the sns color palette
        )
        
    # Add variable labels to the area under the line plot, if requested
    if label_variable_on_area:
        for variable in dataframe[variable_column_name].unique():
            # Get the data for the variable
            variable_data = dataframe[dataframe[variable_column_name] == variable]
            
            # Filter to records where the value is greater than 0
            variable_data = variable_data[variable_data[value_column_name] > 0]
        
            # Get the maximum value for the variable
            max_value = variable_data[value_column_name].max()
            
            # Add the variable label to the area under the line plot
            ax.text(
                x=variable_data[time_column_name].iloc[int(len(variable_data[time_column_name])/2)],
                y=max_value*0.4,
                s=variable,
                verticalalignment="bottom",
                # fontname="Arial",
                fontsize=label_font_size,
                fontweight="bold",
                color=label_font_color # Get the color of the last line in the plot
            )
    
    # Remove top and right spines, and set bottom and left spines to gray
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_color('#666666')
    ax.spines['left'].set_color('#666666')
    
    # Set the number of ticks on the x-axis
    if number_of_x_axis_ticks is not None:
        ax.xaxis.set_major_locator(plt.MaxNLocator(number_of_x_axis_ticks))
        
    # Rotate x-axis tick labels
    if x_axis_tick_rotation is not None:
        plt.xticks(rotation=x_axis_tick_rotation)

    # Format tick labels to be Arial, size 9, and color #666666
    ax.tick_params(
        which='major',
        labelsize=9,
        color='#666666'
    )
    
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Word wrap the subtitle without splitting words
    if subtitle_for_plot != None:   
        subtitle_for_plot = textwrap.fill(subtitle_for_plot, 110, break_long_words=False)
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            # fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
    
    # Move the y-axis label to the top of the y-axis, and set the font to Arial, size 9, and color #666666
    ax.yaxis.set_label_coords(-0.1, 0.84)
    ax.yaxis.set_label_text(
        value_column_name,
        # fontname="Arial",
        fontsize=10,
        color="#666666"
    )
    
    # Move the x-axis label to the right of the x-axis, and set the font to Arial, size 9, and color #666666
    ax.xaxis.set_label_coords(0.9, -0.1)
    ax.xaxis.set_label_text(
        time_column_name,
        # fontname="Arial",
        fontsize=10,
        color="#666666"
    )
    
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )

    # Show plot
    plt.show()
    



# Load packages
import pandas as pd
import numpy as np

# Declare function
def AddLeadingZeros(dataframe,
                    column_name,
                    fixed_length=None,
                    add_as_new_column=False):
    """
    This function adds leading zeros to a column. If fixed_length is not specified, the longest string in the column is used as the fixed length. If add_as_new_column is set to True, the new column is added to the dataframe. Otherwise, the original column is updated.
    
    Args:
        dataframe (Pandas dataframe): Pandas dataframe
        column_name (str): Name of column to add leading zeros to
        fixed_length (int, optional): The length each value in the column should be. Defaults to None.
        add_as_new_column (bool, optional): Whether the updated values with leading zeros should be added as a new column. Defaults to False.

    Returns:
        Pandas dataframe: An updated Pandas dataframe with leading zeros added to the specified column.
    """
    
    # If fixed length not specified, set the longest string as the fixed length
    if fixed_length == None:
        fixed_length = max(dataframe[column_name].astype(str).str.len())
    
    # If adding as new column, change the column name
    if add_as_new_column:
        new_column_name = column_name + ' - with leading 0s'
        dataframe[new_column_name] = np.where(
            dataframe[column_name].isna(),
            np.nan,
            dataframe[column_name].astype(str).str.zfill(fixed_length)
        )
        dataframe[new_column_name] = dataframe[new_column_name].str.replace(".0", "",
                                                                            regex=False)
    else:
        dataframe[column_name] = np.where(
            dataframe[column_name].isna(),
            np.nan,
            dataframe[column_name].astype(str).str.zfill(fixed_length)
        )
        dataframe[column_name] = dataframe[column_name].str.replace(".0", "",
                                                                    regex=False)
    
    # Return updated dataframe
    return(dataframe)




from .PlotBarChart import PlotBarChart
from .PlotBoxWhiskerByGroup import PlotBoxWhiskerByGroup
from .PlotBulletChart import PlotBulletChart
from .PlotClusteredBarChart import PlotClusteredBarChart
from .PlotContingencyHeatmap import PlotContingencyHeatmap
from .PlotCorrelationMatrix import PlotCorrelationMatrix
from .PlotDensityByGroup import PlotDensityByGroup
from .PlotDotPlot import PlotDotPlot
from .PlotHeatmap import PlotHeatmap
from .PlotOverlappingAreaChart import PlotOverlappingAreaChart
from .PlotRiskTolerance import PlotRiskTolerance
from .PlotScatterplot import PlotScatterplot
from .PlotSingleVariableCountPlot import PlotSingleVariableCountPlot
from .PlotSingleVariableHistogram import PlotSingleVariableHistogram
from .PlotTimeSeries import PlotTimeSeries
from .RenderTableOne import RenderTableOne


# Load packages
from dotenv import load_dotenv
from IPython.display import display, HTML, Markdown
import json
import os
import requests

# Declare function
def GetGoogleSearchResults(query,
                           serper_api_key=None,
                           number_of_results=10,
                           apply_autocorrect=False,
                           display_results=False):
    """
    Fetches Google search results for a given query using the Serper API.

    Args:
        query (str): The search query to get results for.
        serper_api_key (str, optional): The API key for the Serper service. If not provided, it will try to load from .env file. Defaults to None.
        number_of_results (int, optional): The number of search results to return. Defaults to 10.
        apply_autocorrect (bool, optional): Whether to apply autocorrect to the query. Defaults to False.
        display_results (bool, optional): Whether to display the results in a DataFrame. Defaults to False.

    Raises:
        ValueError: If no API key is provided and no .env file is found.

    Returns:
        dict: The response from the Serper API, containing the search results.
    """
    
    # If no API key is provided, load from .env file
    if serper_api_key is None:
        load_dotenv()
        try:
            serper_api_key = os.environ['SERPER_API_KEY']
        except:
            raise ValueError("No API key provided and no .env file found. If you need a Serper API key, visit https://serper.dev/")
    
    # Set Google Search URL
    url = "https://google.serper.dev/search"
    
    # Set payload to query
    payload = json.dumps({
        'q': query,
        "autocorrect": apply_autocorrect,
        "num": number_of_results
    })
    
    # Define headers for API call
    headers = {
        'X-API-KEY': serper_api_key,
        'Content-Type': 'application/json'
    }
    
    # Get the response from the API
    response = requests.request(
        method="POST", 
        url=url, 
        headers=headers, 
        data=payload
    )
    
    # Convert the response to json
    response = response.json()
    
    # Display the results if requested
    if display_results:
        # Create a dataframe of the organic results
        results_df = pd.DataFrame(results['organic'])
        # Display the dataframe
        display(results_df)
        
    # Return the response
    return response




from .ConductManifoldLearning import ConductManifoldLearning
from .ConductPrincipalComponentAnalysis import ConductPrincipalComponentAnalysis
from .CreateAssociationRules import CreateAssociationRules
from .CreateGaussianMixtureClusters import CreateGaussianMixtureClusters
from .CreateHierarchicalClusters import CreateHierarchicalClusters
from .CreateKMeansClusters import CreateKMeansClusters
from .GenerateEDAWithLIDA import GenerateEDAWithLIDA


# Load packages
import folium 
import geopandas as gp
import mapclassify
import matplotlib.pyplot as plt 
import pygris

# Declare function
def FetchUSShapefile(state=None,
                     county=None,
                     geography='census tract',
                     census_year=2021):
    """
    Fetches a geographical shapefile from the TIGER database of the U.S. Census Bureau.

    Parameters:
        state (str, optional): The state for which the shapefile is to be fetched. Default is None.
        county (str, optional): The county for which the shapefile is to be fetched. Default is None.
        geography (str, optional): The type of geographical unit for which the shapefile is to be fetched. 
                                Must be one of the following: 'nation', 'divisions', 'regions', 'states', 
                                'counties', 'zipcodes', 'tract', 'block groups', 'blocks', 'school districts', 
                                'county subdivisions', 'congressional districts', 'state legislative districts', 
                                'voting districts'. Default is 'census tract'.
        census_year (int, optional): The census year for which the shapefile is to be fetched. Default is 2021.

    Returns:
        geopandas.GeoDataFrame: The fetched shapefile as a GeoDataFrame.
    """
    
    # Ensure that geography is a valid option
    if geography not in ['nation', 'divisions', 'regions', 'states', 'counties', 'zipcodes', 'tract', 'block groups', 'blocks', 'school districts', 'county subdivisions', 'congressional districts', 'state legislative districts', 'voting districts']:
        raise ValueError('geography must be one of the following: nation, divisions, regions, states, counties, zipcodes, tract, block groups, blocks, school districts, county subdivisions, congressional districts, state legislative districts, voting districts')
    
    # Ensure that geography is a valid option
    valid_geographies = {
        'nation': pygris.nation,
        'divisions': pygris.divisions,
        'regions': pygris.regions,
        'states': pygris.states,
        'counties': pygris.counties,
        'zipcodes': pygris.zctas,
        'tract': pygris.tracts,
        'block groups': pygris.block_groups,
        'blocks': pygris.blocks,
        'school districts': pygris.school_districts,
        'county subdivisions': pygris.county_subdivisions,
        'congressional districts': pygris.congressional_districts,
        'state legislative districts': pygris.state_legislative_districts,
        'voting districts': pygris.voting_districts
    }

    if geography not in valid_geographies:
        raise ValueError('geography must be one of the following: ' + ', '.join(valid_geographies.keys()))
    
    # Fetch shapefile from TIGER database from U.S. Census Bureau
    shapefile_function = valid_geographies[geography]
    if geography in ['nation', 'divisions', 'regions', 'states']:
        shapefile = shapefile_function(year=census_year)
    elif geography in ['counties', 'zipcodes']:
        shapefile = shapefile_function(state=state, year=census_year)
    else:
        shapefile = shapefile_function(state=state, county=county, year=census_year)
    
    # Return the shapefile
    return(shapefile)

